import json

import dask.array as da
import pytest

from numpydantic.interface import DaskInterface
from numpydantic.testing.interfaces import DaskCase

pytestmark = pytest.mark.dask


def test_dask_enabled():
    """
    We need dask to be available to run these tests :)
    """
    assert DaskInterface.enabled()


def test_dask_check(interface_cases, tmp_output_dir_func):
    array = interface_cases.make_array(path=tmp_output_dir_func)

    if interface_cases.interface is DaskInterface:
        assert DaskInterface.check(array)
    else:
        assert not DaskInterface.check(array)


@pytest.mark.shape
def test_dask_shape(shape_cases):
    shape_cases.interface = DaskCase
    shape_cases.validate_case()


@pytest.mark.dtype
def test_dask_dtype(dtype_cases):
    dtype_cases.interface = DaskCase
    dtype_cases.validate_case()


@pytest.mark.serialization
def test_dask_to_json(array_model):
    array_list = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    array = da.array(array_list)
    model = array_model((3, 3), int)
    instance = model(array=array)
    jsonified = json.loads(instance.model_dump_json())
    assert jsonified["array"] == array_list



import json
from datetime import datetime
from typing import Any

import h5py
import numpy as np
import pytest
from pydantic import BaseModel

from numpydantic import NDArray, Shape
from numpydantic.interface import H5Interface
from numpydantic.interface.hdf5 import H5ArrayPath, H5Proxy
from numpydantic.testing.interfaces import HDF5Case, HDF5CompoundCase

pytestmark = pytest.mark.hdf5


@pytest.fixture(
    params=[
        pytest.param(HDF5Case, id="hdf5"),
        pytest.param(HDF5CompoundCase, id="hdf5-compound"),
    ]
)
def hdf5_cases(request):
    return request.param


def test_hdf5_enabled():
    assert H5Interface.enabled()


@pytest.mark.shape
def test_hdf5_shape(shape_cases, hdf5_cases):
    shape_cases.interface = hdf5_cases
    if shape_cases.skip():
        pytest.skip()
    shape_cases.validate_case()


@pytest.mark.dtype
def test_hdf5_dtype(dtype_cases, hdf5_cases):
    dtype_cases.interface = hdf5_cases
    dtype_cases.validate_case()


def test_hdf5_check(interface_cases, tmp_output_dir_func):
    array = interface_cases.make_array(path=tmp_output_dir_func)
    if interface_cases.interface is H5Interface:
        assert H5Interface.check(array)
    else:
        assert not H5Interface.check(array)


def test_hdf5_check_not_exists():
    """We should fail a check for a nonexistent hdf5 file"""
    spec = ("./fakefile.h5", "/fake/array")
    assert not H5Interface.check(spec)


def test_hdf5_check_not_hdf5(tmp_path):
    """Files that exist but aren't actually hdf5 files should fail a check"""
    afile = tmp_path / "not_an_hdf.h5"
    with open(afile, "w") as af:
        af.write("hey")

    spec = (afile, "/fake/array")
    assert not H5Interface.check(spec)


def test_hdf5_dataset_not_exists(hdf5_array, model_blank):
    array = hdf5_array()
    with pytest.raises(ValueError) as e:
        model_blank(array=H5ArrayPath(file=array.file, path="/some/random/path"))
        assert "file located" in e
        assert "no array found" in e


@pytest.mark.proxy
def test_assignment(hdf5_array, model_blank):
    array = hdf5_array()

    model = model_blank(array=array)
    model.array[1, 1] = 5
    assert model.array[1, 1] == 5

    model.array[1:3, 2:4] = 10
    assert (model.array[1:3, 2:4] == 10).all()


@pytest.mark.serialization
@pytest.mark.parametrize("round_trip", (True, False))
def test_to_json(hdf5_array, array_model, round_trip):
    """
    Test serialization of HDF5 arrays to JSON
    Args:
        hdf5_array:

    Returns:

    """
    array = hdf5_array((10, 10), int)
    model = array_model((10, 10), int)

    instance = model(array=array)  # type: BaseModel

    json_str = instance.model_dump_json(
        round_trip=round_trip, context={"absolute_paths": True}
    )
    json_dumped = json.loads(json_str)["array"]
    if round_trip:
        assert json_dumped["file"] == str(array.file)
        assert json_dumped["path"] == str(array.path)
    else:
        assert json_dumped == instance.array[:].tolist()


@pytest.mark.dtype
@pytest.mark.proxy
def test_compound_dtype(tmp_path):
    """
    hdf5 proxy indexes compound dtypes as single fields when field is given
    """
    h5f_path = tmp_path / "test.h5"
    dataset_path = "/dataset"
    field = "data"
    dtype = np.dtype([(field, "i8"), ("extra", "f8")])
    data = np.zeros((10, 20), dtype=dtype)
    with h5py.File(h5f_path, "w") as h5f:
        dset = h5f.create_dataset(dataset_path, data=data)
        assert dset.dtype == dtype

    proxy = H5Proxy(h5f_path, dataset_path, field=field)
    assert proxy.dtype == np.dtype("int64")
    assert proxy.shape == (10, 20)
    assert proxy[0, 0] == 0

    class MyModel(BaseModel):
        array: NDArray[Shape["10, 20"], np.int64]

    instance = MyModel(array=(h5f_path, dataset_path, field))
    assert instance.array.dtype == np.dtype("int64")
    assert instance.array.shape == (10, 20)
    assert instance.array[0, 0] == 0

    # set values too
    instance.array[0, :] = 1
    assert all(instance.array[0, :] == 1)
    assert all(instance.array[1, :] == 0)
    instance.array[1] = 2
    assert all(instance.array[1] == 2)


@pytest.mark.dtype
@pytest.mark.proxy
@pytest.mark.parametrize("compound", [True, False])
def test_strings(hdf5_array, compound):
    """
    HDF5 proxy can get and set strings just like any other dtype
    """
    array = hdf5_array((10, 10), str, compound=compound)

    class MyModel(BaseModel):
        array: NDArray[Shape["10, 10"], str]

    instance = MyModel(array=array)
    instance.array[0, 0] = "hey"
    assert instance.array[0, 0] == "hey"
    assert isinstance(instance.array[0, 1], str)

    instance.array[1] = "sup"
    assert all(instance.array[1] == "sup")


@pytest.mark.dtype
@pytest.mark.proxy
@pytest.mark.parametrize("compound", [True, False])
def test_datetime(hdf5_array, compound):
    """
    We can treat S32 byte arrays as datetimes if our type annotation
    says to, including validation, setting and getting values
    """
    array = hdf5_array((10, 10), datetime, compound=compound)

    class MyModel(BaseModel):
        array: NDArray[Any, datetime]

    instance = MyModel(array=array)
    assert isinstance(instance.array[0, 0], np.datetime64)
    assert instance.array[0:5].dtype.type is np.datetime64

    now = datetime.now()

    instance.array[0, 0] = now
    assert instance.array[0, 0] == now
    instance.array[0] = now
    assert all(instance.array[0] == now)


@pytest.mark.parametrize("dtype", [int, float, str, datetime])
def test_empty_dataset(dtype, tmp_path):
    """
    Empty datasets shouldn't choke us during validation
    """
    array_path = tmp_path / "test.h5"
    np_dtype = "S32" if dtype in (str, datetime) else dtype

    with h5py.File(array_path, "w") as h5f:
        _ = h5f.create_dataset(name="/data", dtype=np_dtype)

    class MyModel(BaseModel):
        array: NDArray[Any, dtype]

    _ = MyModel(array=(array_path, "/data"))


@pytest.mark.proxy
@pytest.mark.parametrize(
    "comparison,valid",
    [
        (H5Proxy(file="test_file.h5", path="/subpath", field="sup"), True),
        (H5Proxy(file="test_file.h5", path="/subpath"), False),
        (H5Proxy(file="different_file.h5", path="/subpath"), False),
        (("different_file.h5", "/subpath", "sup"), ValueError),
        ("not even a proxy-like thing", ValueError),
    ],
)
def test_proxy_eq(comparison, valid):
    """
    test the __eq__ method of H5ArrayProxy matches proxies to the same
    dataset (and path), or raises a ValueError
    """
    proxy_a = H5Proxy(file="test_file.h5", path="/subpath", field="sup")
    if valid is True:
        assert proxy_a == comparison
    elif valid is False:
        assert proxy_a != comparison
    else:
        with pytest.raises(valid):
            assert proxy_a == comparison



import json
from datetime import datetime
from typing import Any

import numpy as np
import pytest
from pydantic import BaseModel

from numpydantic import NDArray, Shape
from numpydantic.testing.cases import NumpyCase

pytestmark = pytest.mark.numpy


@pytest.mark.shape
def test_numpy_shape(shape_cases):
    shape_cases.interface = NumpyCase
    shape_cases.validate_case()


@pytest.mark.dtype
def test_numpy_dtype(dtype_cases):
    dtype_cases.interface = NumpyCase
    dtype_cases.validate_case()


@pytest.mark.dtype
def test_numpy_datetime64():
    """
    Special testing for datetime64, since it's a numpy-specific thing,
    and requires special handling in the other interfaces.

    Typically, we expect that generic type annotation to be a `datetime.datetime`,
    but it should also be possible to use a numpy.datetime64 type when doing
    numpy validation.

    So this test is just the simplest regression test to make sure we can use datetime64
    """
    arr = np.array([np.datetime64(datetime.now())], dtype=np.datetime64)

    annotation = NDArray[Shape["*"], np.datetime64]
    _ = annotation(arr)

    class MyModel(BaseModel):
        array: annotation

    _ = MyModel(array=arr)


def test_numpy_coercion(model_blank):
    """If no other interface matches, we try and coerce to a numpy array"""
    instance = model_blank(array=[1, 2, 3])
    assert isinstance(instance.array, np.ndarray)


@pytest.mark.scalar
@pytest.mark.serialization
def test_numpy_empty_string():
    """
    Empty strings are coerced to arrays and serialzied as such.

    Mildly redundant with test_serialize_scalars_as_arrays below,
    but a specific regression test for issue #52

    References:
        - https://github.com/p2p-ld/numpydantic/issues/52
    """

    class MyModel(BaseModel):
        array: NDArray[Any, np.str_]

    inst = MyModel(array="")
    assert isinstance(inst.array, np.ndarray)
    assert inst.array == np.array([""])
    dumped = inst.model_dump_json()
    assert json.loads(dumped)["array"] == [""]


@pytest.mark.serialization
@pytest.mark.scalar
@pytest.mark.parametrize("scalar", ("", 0, 0.5))
@pytest.mark.parametrize("as_json", (True, False))
def test_serialize_scalars_as_arrays(scalar, as_json: bool):
    """
    The numpy interface matches scalar values, coerces them to arrays,
    and serializes them as arrays in both python and json.
    """

    class MyModel(BaseModel):
        array: NDArray

    inst = MyModel(array=scalar)
    assert isinstance(inst.array, np.ndarray)
    assert inst.array == np.array([scalar])
    if as_json:
        dumped = inst.model_dump_json()
        assert json.loads(dumped)["array"] == [scalar]
    else:
        dumped = inst.model_dump()
        assert dumped["array"] == np.array([scalar])



"""
Interface to zarr arrays
"""

import contextlib
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal, Optional, Sequence, Union

import numpy as np
from pydantic import SerializationInfo

from numpydantic.interface.interface import Interface, JsonDict
from numpydantic.types import DtypeType

try:
    import zarr
    from numcodecs import VLenUTF8
    from zarr.core import Array as ZarrArray
    from zarr.storage import StoreLike
except ImportError:  # pragma: no cover
    ZarrArray = None
    StoreLike = None
    storage = None
    VLenUTF8 = None


@dataclass
class ZarrArrayPath:
    """
    Map to an array within a zarr store.

    See :func:`zarr.open`
    """

    file: Union[Path, str]
    """Location of Zarr store file or directory"""
    path: Optional[str] = None
    """Path to array within hierarchical zarr store"""

    def open(self, **kwargs: dict) -> ZarrArray:
        """Open the zarr array at the provided path"""
        return zarr.open(str(self.file), path=self.path, **kwargs)

    @classmethod
    def from_iterable(cls, spec: Sequence) -> "ZarrArrayPath":
        """
        Construct a :class:`.ZarrArrayPath` specifier from an iterable,
        rather than kwargs
        """
        if len(spec) == 1:
            return ZarrArrayPath(file=spec[0])
        elif len(spec) == 2:
            return ZarrArrayPath(file=spec[0], path=spec[1])
        else:
            raise ValueError("Only len 1-2 iterables can be used for a ZarrArrayPath")


class ZarrJsonDict(JsonDict):
    """Round-trip Json-able version of a Zarr Array"""

    info: dict[str, str]
    type: Literal["zarr"]
    file: Optional[str] = None
    path: Optional[str] = None
    dtype: Optional[str] = None
    value: Optional[list] = None

    def to_array_input(self) -> Union[ZarrArray, ZarrArrayPath]:
        """
        Construct a ZarrArrayPath if file and path are present,
        otherwise a ZarrArray
        """
        if self.file:
            array = ZarrArrayPath(file=self.file, path=self.path)
        else:
            dtype = np.str_ if self.dtype == "str" else self.dtype
            array = zarr.array(self.value, dtype=dtype)
        return array


class ZarrInterface(Interface):
    """
    Interface to in-memory or on-disk zarr arrays
    """

    name = "zarr"
    input_types = (Path, ZarrArray, ZarrArrayPath)
    return_type = ZarrArray
    json_model = ZarrJsonDict

    @classmethod
    def enabled(cls) -> bool:
        """True if zarr is installed"""
        return ZarrArray is not None

    @staticmethod
    def _get_array(
        array: Union[ZarrArray, str, dict, ZarrJsonDict, Path, ZarrArrayPath, Sequence],
    ) -> ZarrArray:
        if isinstance(array, ZarrArray):
            return array

        if isinstance(array, (str, Path)):
            array = ZarrArrayPath(file=array)
        elif isinstance(array, (tuple, list)):
            array = ZarrArrayPath.from_iterable(array)

        return array.open(mode="a")

    @classmethod
    def check(cls, array: Any) -> bool:
        """
        Check if array is in-memory zarr array,
        a path to a zarr array, or a :class:`.ZarrArrayPath`
        """
        if isinstance(array, ZarrArray):
            return True

        if isinstance(array, dict):
            if array.get("type", False) == cls.name:
                return True
            # continue checking if dict contains a zarr file
            array = array.get("file", "")

        # See if can be coerced to ZarrArrayPath
        if isinstance(array, (Path, str)):
            array = ZarrArrayPath(file=array)

        if isinstance(array, (tuple, list)):
            # something that can be coerced to ZarrArrayPath
            with contextlib.suppress(ValueError):
                array = ZarrArrayPath.from_iterable(array)

        if isinstance(array, ZarrArrayPath):
            with contextlib.suppress(Exception):
                arr = array.open(mode="r")
                if isinstance(arr, ZarrArray):
                    return True

        return False

    def before_validation(
        self, array: Union[ZarrArray, str, Path, ZarrArrayPath, Sequence]
    ) -> ZarrArray:
        """
        Ensure that the zarr array is opened
        """
        return self._get_array(array)

    def get_dtype(self, array: ZarrArray) -> DtypeType:
        """
        Override base dtype getter to handle zarr's string-as-object encoding.
        """
        if (
            getattr(array.dtype, "type", None) is np.object_
            and array.filters
            and any([isinstance(f, VLenUTF8) for f in array.filters])
        ):
            return np.str_
        else:
            return array.dtype

    @classmethod
    def to_json(
        cls,
        array: Union[ZarrArray, str, Path, ZarrArrayPath, Sequence],
        info: Optional[SerializationInfo] = None,
    ) -> Union[list, ZarrJsonDict]:
        """
        Dump a Zarr Array to JSON

        If ``info.round_trip == False``, dump the array as a list of lists.
        This may be a memory-intensive operation.

        Otherwise, dump the metadata for an array from
        :meth:`zarr.core.Array.info_items`
        plus the :meth:`zarr.core.Array.hexdigest` as a :class:`.ZarrJsonDict`

        If either the ``dump_array`` value in the context dictionary is ``True``
        or the zarr array is an in-memory array, dump the array as well
        (since without a persistent array it would be impossible to roundtrip and
        dumping to JSON would be meaningless)

        Passing ```dump_array': True`` to the serialization ``context``
        looks like this::

            model.model_dump_json(context={'zarr_dump_array': True})
        """
        array = cls._get_array(array)

        if info.round_trip:
            dump_array = False
            if info is not None and info.context is not None:
                dump_array = info.context.get("dump_array", False)
            is_file = False

            as_json = {"type": cls.name}
            as_json["dtype"] = array.dtype.name
            if hasattr(array.store, "dir_path"):
                is_file = True
                as_json["file"] = array.store.dir_path()
                as_json["path"] = array.name
            as_json["info"] = {i[0]: i[1] for i in array.info_items()}
            as_json["info"]["hexdigest"] = array.hexdigest()

            if dump_array or not is_file:
                as_json["value"] = array[:].tolist()

            as_json = ZarrJsonDict(**as_json)
        else:
            as_json = array[:].tolist()

        return as_json



from pathlib import Path
from typing import Callable, Tuple, Union

import numpy as np
import pytest
import zarr

from numpydantic.interface.hdf5 import H5ArrayPath
from numpydantic.interface.zarr import ZarrArrayPath
from numpydantic.testing.interfaces import HDF5Case, HDF5CompoundCase, VideoCase


@pytest.fixture(scope="function")
def hdf5_array(
    request, tmp_output_dir_func
) -> Callable[[Tuple[int, ...], Union[np.dtype, type]], H5ArrayPath]:

    def _hdf5_array(
        shape: Tuple[int, ...] = (10, 10),
        dtype: Union[np.dtype, type] = float,
        compound: bool = False,
    ) -> H5ArrayPath:
        if compound:
            array: H5ArrayPath = HDF5CompoundCase.make_array(
                shape, dtype, tmp_output_dir_func
            )
            return array
        else:
            return HDF5Case.make_array(shape, dtype, tmp_output_dir_func)

    return _hdf5_array


@pytest.fixture(scope="function")
def zarr_nested_array(tmp_output_dir_func) -> ZarrArrayPath:
    """Zarr array within a nested array"""
    file = tmp_output_dir_func / "nested.zarr"
    path = "a/b/c"
    root = zarr.open(str(file), mode="w")
    _ = root.zeros(path, shape=(100, 100), chunks=(10, 10))
    return ZarrArrayPath(file=file, path=path)


@pytest.fixture(scope="function")
def zarr_array(tmp_output_dir_func) -> Path:
    file = tmp_output_dir_func / "array.zarr"
    array = zarr.open(str(file), mode="w", shape=(100, 100), chunks=(10, 10))
    array[:] = 0
    return file


@pytest.fixture(scope="function")
def avi_video(tmp_output_dir_func) -> Callable[[Tuple[int, int], int, bool], Path]:

    def _make_video(shape=(100, 50), frames=10, is_color=True) -> Path:
        shape = (frames, *shape)
        if is_color:
            shape = (*shape, 3)
        return VideoCase.make_array(
            shape=shape, dtype=np.uint8, path=tmp_output_dir_func
        )

    return _make_video



"""
Helper functions for validation
"""

from numpydantic.validation.dtype import validate_dtype
from numpydantic.validation.shape import validate_shape

__all__ = [
    "validate_dtype",
    "validate_shape",
]



import pytest

from numpydantic.testing.cases import DTYPE_CASES, SHAPE_CASES
from numpydantic.testing.helpers import ValidationCase
from tests.fixtures import *


def pytest_addoption(parser):
    parser.addoption(
        "--with-output",
        action="store_true",
        help="Keep test outputs in the __tmp__ directory",
    )


@pytest.fixture(
    scope="function",
    params=[pytest.param(c, id=c.id, marks=c.pytest_marks) for c in SHAPE_CASES],
)
def shape_cases(request, tmp_output_dir_func) -> ValidationCase:
    case: ValidationCase = request.param.model_copy()
    case.path = tmp_output_dir_func
    return case


@pytest.fixture(
    scope="function",
    params=[pytest.param(c, id=c.id, marks=c.pytest_marks) for c in DTYPE_CASES],
)
def dtype_cases(request, tmp_output_dir_func) -> ValidationCase:
    case: ValidationCase = request.param.model_copy()
    case.path = tmp_output_dir_func
    return case



"""
Needs to be refactored to DRY, but works for now
"""

from pathlib import Path

import cv2
import pytest
from pydantic import BaseModel, ValidationError

from numpydantic import NDArray, Shape
from numpydantic import dtype as dt
from numpydantic.interface.video import VideoProxy

pytestmark = pytest.mark.video


@pytest.mark.parametrize("input_type", [str, Path])
def test_video_validation(avi_video, input_type):
    """Color videos should validate for normal uint8 shape specs"""

    shape = (100, 50)
    vid = avi_video(shape=shape, is_color=True)
    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    # should correctly validate :)
    instance = MyModel(array=input_type(vid))
    assert isinstance(instance.array, VideoProxy)


def test_video_from_videocapture(avi_video):
    """Should be able to pass an opened videocapture object"""
    shape = (100, 50)
    vid = avi_video(shape=shape, is_color=True)
    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    # should still correctly validate!
    opened_vid = cv2.VideoCapture(str(vid))
    try:
        instance = MyModel(array=opened_vid)
        assert isinstance(instance.array, VideoProxy)
    finally:
        opened_vid.release()


@pytest.mark.shape
def test_video_wrong_shape(avi_video):
    shape = (100, 50)

    # generate video with purposely wrong shape
    vid = avi_video(shape=(shape[0] + 10, shape[1] + 10), is_color=True)

    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    # should correctly validate :)
    with pytest.raises(ValidationError):
        _ = MyModel(array=vid)


@pytest.mark.proxy
def test_video_getitem(avi_video):
    """
    Should be able to get individual frames and slices as if it were a normal array
    """
    shape = (100, 50)
    vid = avi_video(shape=shape, frames=10, is_color=True)
    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    instance = MyModel(array=vid)
    fifth_frame = instance.array[5]
    # the fifth frame should be all 5s
    assert (fifth_frame[5, 5, :] == [5, 5, 5]).all()

    # slicing should also work as if it were just a numpy array
    single_slice = instance.array[3, 0:10, 0:5]
    assert single_slice[3, 3, 0] == 3
    assert single_slice.shape == (10, 5, 3)

    # also get a range of frames
    # range without further slices
    range_slice = instance.array[3:5]
    assert range_slice.shape == (2, 100, 50, 3)
    assert range_slice[0, 3, 3, 0] == 3
    assert range_slice[1, 4, 4, 0] == 4

    # full range
    range_slice = instance.array[3:5, 0:10, 0:5]
    assert range_slice.shape == (2, 10, 5, 3)
    assert range_slice[0, 3, 3, 0] == 3
    assert range_slice[1, 4, 4, 0] == 4

    # starting range
    range_slice = instance.array[6:, 0:10, 0:10]
    assert range_slice.shape == (4, 10, 10, 3)
    assert range_slice[-1, 9, 9, 0] == 9
    assert range_slice[-2, 9, 9, 0] == 8

    # ending range
    range_slice = instance.array[:3, 0:5, 0:5]
    assert range_slice.shape == (3, 5, 5, 3)

    # stepped range
    range_slice = instance.array[0:5:2, 0:6, 0:6]
    # second slice should be the second frame (instead of the first)
    assert range_slice.shape == (3, 6, 6, 3)
    assert range_slice[1, 2, 2, 0] == 2
    # and the third should be the fourth (instead of the second)
    assert range_slice[2, 4, 4, 0] == 4

    with pytest.raises(NotImplementedError):
        # shouldn't be allowed to set
        instance.array[5] = 10


@pytest.mark.proxy
def test_video_attrs(avi_video):
    """Should be able to access opencv properties"""
    shape = (100, 50)
    vid = avi_video(shape=shape, is_color=True)
    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    instance = MyModel(array=vid)

    instance.array.set(cv2.CAP_PROP_POS_FRAMES, 5)
    assert int(instance.array.get(cv2.CAP_PROP_POS_FRAMES)) == 5


@pytest.mark.proxy
def test_video_close(avi_video):
    """Should close and reopen video file if needed"""
    shape = (100, 50)
    vid = avi_video(shape=shape, is_color=True)
    shape_str = f"*, {shape[0]}, {shape[1]}, 3"

    class MyModel(BaseModel):
        array: NDArray[Shape[shape_str], dt.UInt8]

    instance = MyModel(array=vid)
    assert isinstance(instance.array.video, cv2.VideoCapture)
    # closes releases and removed reference
    instance.array.close()
    assert instance.array._video is None
    # reopen
    assert isinstance(instance.array.video, cv2.VideoCapture)


@pytest.mark.proxy
def test_video_not_exists(tmp_path):
    """
    A video file that doesn't exist should raise an error
    """
    video = VideoProxy(tmp_path / "not_real.avi")
    with pytest.raises(FileNotFoundError):
        _ = video.video


@pytest.mark.proxy
@pytest.mark.parametrize(
    "comparison,valid",
    [
        (VideoProxy("test_video.avi"), True),
        (VideoProxy("not_real_video.avi"), False),
        ("not even a video proxy", TypeError),
    ],
)
def test_video_proxy_eq(comparison, valid):
    """
    Comparing a video proxy's equality should be valid if the path matches
    Args:
        comparison:
        valid:

    Returns:

    """
    proxy_a = VideoProxy("test_video.avi")
    if valid is True:
        assert proxy_a == comparison
    elif valid is False:
        assert proxy_a != comparison
    else:
        with pytest.raises(valid):
            assert proxy_a == comparison



"""
Helper functions for use with :class:`~numpydantic.NDArray` - see the note in
:mod:`~numpydantic.ndarray` for why these are separated.
"""

import hashlib
import json
from typing import TYPE_CHECKING, Any, Callable, Optional, get_args

import numpy as np
from pydantic import BaseModel
from pydantic_core import CoreSchema, core_schema
from pydantic_core.core_schema import ListSchema, ValidationInfo

from numpydantic import dtype as dt
from numpydantic.interface import Interface
from numpydantic.maps import np_to_python
from numpydantic.types import DtypeType, NDArrayType, ShapeType
from numpydantic.validation.dtype import is_union
from numpydantic.vendor.nptyping.structure import StructureMeta

if TYPE_CHECKING:  # pragma: no cover
    from pydantic._internal._schema_generation_shared import (
        CallbackGetCoreSchemaHandler,
    )

    from numpydantic import Shape


def _numeric_dtype(
    dtype: DtypeType, _handler: "CallbackGetCoreSchemaHandler"
) -> CoreSchema:
    """Make a numeric dtype that respects min/max values from extended numpy types"""
    if dtype in (np.number,):
        dtype = float

    if issubclass(dtype, np.floating):
        info = np.finfo(dtype)
        schema = core_schema.float_schema(le=float(info.max), ge=float(info.min))
    elif issubclass(dtype, np.integer):
        info = np.iinfo(dtype)
        schema = core_schema.int_schema(le=int(info.max), ge=int(info.min))
    else:
        schema = _handler.generate_schema(dtype)

    return schema


def _lol_dtype(
    dtype: DtypeType, _handler: "CallbackGetCoreSchemaHandler"
) -> CoreSchema:
    """Get the innermost dtype schema to use in the generated pydantic schema"""
    if isinstance(dtype, StructureMeta):  # pragma: no cover
        raise NotImplementedError("Structured dtypes are currently unsupported")
    if isinstance(dtype, tuple):
        # if it's a meta-type that refers to a generic float/int, just make that
        if dtype in (dt.Float, dt.Number):
            array_type = core_schema.float_schema()
        elif dtype == dt.Integer:
            array_type = core_schema.int_schema()
        elif dtype == dt.Complex:
            array_type = core_schema.any_schema()
        else:
            # make a union of dtypes recursively
            types_ = list(set(dtype))
            array_type = core_schema.union_schema(
                [_lol_dtype(t, _handler) for t in types_]
            )
    elif is_union(dtype):
        array_type = core_schema.union_schema(
            [_lol_dtype(t, _handler) for t in get_args(dtype)]
        )
    else:
        try:
            python_type = np_to_python[dtype]
        except KeyError:  # pragma: no cover
            # this should pretty much only happen in downstream/3rd-party interfaces
            # that use interface-specific types. those need to provide mappings back
            # to base python types (making this more streamlined is TODO)
            if dtype in np_to_python.values():
                # it's already a python type
                python_type = dtype
            elif issubclass(dtype, BaseModel):
                python_type = dtype
            else:
                # does this need a warning?
                python_type = Any

        if python_type in (float, int):
            array_type = _numeric_dtype(dtype, _handler)
        elif python_type is bool:
            array_type = core_schema.bool_schema()
        elif python_type is Any:
            array_type = core_schema.any_schema()
        elif issubclass(python_type, BaseModel):
            # We a) need to materialize the model schema rather than reference it
            # since there isn't a good way to propagate the reference through the
            # handler,
            # and b) can't use the handler to generate the schema
            # since it can only generate the reference
            # caching it for full materialization later.
            # the easiest way to get the model's schema is just to ... get it,
            # which may be a bit expensive with the copy.
            # see https://github.com/p2p-ld/numpydantic/issues/48
            # if this proves to be prohibitively expensive, we can find a better way.
            array_type = python_type.__pydantic_core_schema__.copy()
        else:
            array_type = _handler.generate_schema(python_type)

    return array_type


def list_of_lists_schema(shape: "Shape", array_type: CoreSchema) -> ListSchema:
    """
    Make a pydantic JSON schema for an array as a list of lists.

    For each item in the shape, create a list schema. In the innermost schema
    insert the passed ``array_type`` schema.

    This function is typically called from :func:`.make_json_schema`

    Args:
        shape (:class:`~numpydantic.Shape`): Shape determines the depth and max/min
            elements for each layer of list schema
        array_type ( :class:`pydantic_core.CoreSchema` ): The pre-rendered pydantic
            core schema to use in the innermost list entry
    """
    from numpydantic.validation.shape import _is_range

    shape_parts = [part.strip() for part in shape.__args__[0].split(",")]
    # labels, if present
    split_parts = [
        p.split(" ")[1] if len(p.split(" ")) == 2 else None for p in shape_parts
    ]

    # Construct a list of list schema
    # go in reverse order - construct list schemas such that
    # the final schema is the one that checks the first dimension
    shape_labels = reversed(split_parts)
    shape_args = reversed(shape.prepared_args)
    list_schema = None
    for arg, label in zip(shape_args, shape_labels):
        # which handler to use? for the first we use the actual type
        # handler, everywhere else we use the prior list handler
        inner_schema = array_type if list_schema is None else list_schema

        # make a label annotation, if we have one
        metadata = {"name": label} if label is not None else None

        # make the current level list schema, accounting for shape
        if arg == "*":
            list_schema = core_schema.list_schema(inner_schema, metadata=metadata)
        elif arg == "...":
            list_schema = _unbounded_shape(inner_schema, metadata=metadata)
        else:
            if _is_range(arg):
                arg_min, arg_max = arg.split("-")
                arg_min = None if arg_min == "*" else int(arg_min)
                arg_max = None if arg_max == "*" else int(arg_max)

            else:
                try:
                    arg = int(arg)
                    arg_min = arg
                    arg_max = arg
                except ValueError as e:  # pragma: no cover

                    raise ValueError(
                        "Array shapes must be integers, wildcards, ellipses, or "
                        "ranges. Shape variables (for declaring that one dimension "
                        "must be the same size as another) are not supported because "
                        "it is impossible to express dynamic minItems/maxItems in "
                        "JSON Schema. "
                        "See: https://github.com/orgs/json-schema-org/discussions/730"
                    ) from e
            list_schema = core_schema.list_schema(
                inner_schema, min_length=arg_min, max_length=arg_max, metadata=metadata
            )
    return list_schema


def _hash_schema(schema: CoreSchema) -> str:
    """
    Make a hex-encoded 8-byte blake2b hash from a pydantic core schema.
    Collisions are really not important or likely here, but we do want the same schema
    to produce the same hash.
    """
    schema_str = json.dumps(
        schema,
        sort_keys=True,
        indent=None,
        separators=(",", ":"),
        default=lambda x: None,
    ).encode("utf-8")
    hasher = hashlib.blake2b(digest_size=8)
    hasher.update(schema_str)
    return hasher.hexdigest()


def _unbounded_shape(
    inner_type: CoreSchema, metadata: Optional[dict] = None
) -> core_schema.DefinitionsSchema:
    """
    Make a recursive schema that refers to itself using a hashed version of the inner
    type
    """

    schema_hash = _hash_schema(inner_type)
    array_ref = f"any-shape-array-{schema_hash}"

    schema = core_schema.definitions_schema(
        core_schema.list_schema(
            core_schema.definition_reference_schema(array_ref), metadata=metadata
        ),
        [
            core_schema.union_schema(
                [
                    core_schema.list_schema(
                        core_schema.definition_reference_schema(array_ref)
                    ),
                    inner_type,
                ],
                ref=array_ref,
            )
        ],
    )
    return schema


def make_json_schema(
    shape: ShapeType, dtype: DtypeType, _handler: "CallbackGetCoreSchemaHandler"
) -> ListSchema:
    """
    Make a list of list pydantic core schema for an array from a shape and a dtype.
    Used to generate JSON schema in the containing model, but not for validation,
    which is handled by interfaces.

    First resolves the dtype into a pydantic ``CoreSchema`` ,
    and then uses that with :func:`.list_of_lists_schema` .

    Args:
        shape ( ShapeType ): Specification of a shape, as a tuple or
            an nptyping ``Shape``
        dtype ( DtypeType ): A builtin type or numpy dtype
        _handler: The pydantic schema generation handler (see pydantic docs).
            Needed to generate the schema for the dtype
            if we don't know how to do it ourselves.

    Returns:
        :class:`pydantic_core.core_schema.ListSchema`
    """
    dtype_schema = _lol_dtype(dtype, _handler)

    # get the names of the shape constraints, if any
    if shape is Any:
        list_schema = _unbounded_shape(dtype_schema)
        # list_schema = core_schema.list_schema(core_schema.any_schema())
    else:
        list_schema = list_of_lists_schema(shape, dtype_schema)

    return list_schema


def get_validate_interface(shape: ShapeType, dtype: DtypeType) -> Callable:
    """
    Validate using a matching :class:`.Interface` class using its
    :meth:`.Interface.validate` method
    """

    def validate_interface(
        value: Any, info: Optional["ValidationInfo"] = None
    ) -> NDArrayType:
        interface_cls = Interface.match(value)
        interface = interface_cls(shape, dtype)
        value = interface.validate(value)
        return value

    return validate_interface



import sys

import pytest

from numpydantic import NDArray
from numpydantic.meta import update_ndarray_stub

if sys.version_info.minor < 11:
    from typing_extensions import reveal_type
else:
    from typing import reveal_type


def test_no_warn(recwarn):
    """
    If something is going wrong with generating meta stubs, a warning will be emitted.
    that is bad.
    """
    update_ndarray_stub()
    assert len(recwarn) == 0


@pytest.mark.skip("TODO")
def test_generate_stub():
    """
    Test that we generate the stub file correctly...
    """
    pass


@pytest.mark.skip("TODO")
def test_update_stub():
    """
    Test that the update stub file correctly updates the stub stored in the package
    """
    pass


@pytest.mark.skip("TODO")
def test_stub_revealed_type():
    """
    Check that the revealed type matches the stub
    """
    _ = reveal_type(NDArray)



from typing import Any
from typing import Literal as L

import numpy as np

from numpydantic import NDArray, Shape

x: NDArray[Shape[L["1"]], Any] = np.empty((1,))


def a_func(array: np.typing.NDArray) -> None:
    pass


if isinstance(x, np.ndarray):
    a_func(x)



"""
Vendored modules - see licenses in /licenses.

Currently consists just of nptyping, as it is no longer maintained
and pins a version of numpy<2, and we have to do an increasing
number of awkward monkeypatches for perf and customization reasons.

This vendored module will slowly be worked out of the code and
all its functionality replaced.
"""



"""
Tests that should be applied to all interfaces
"""

import json
from importlib.metadata import version

import dask.array as da
import numpy as np
import pytest
from pydantic import BaseModel
from zarr.core import Array as ZarrArray

from numpydantic.interface import Interface, InterfaceMark, MarkedJson
from numpydantic.testing.cases import ZERO_LENGTH_CASES_PASSING
from numpydantic.testing.helpers import ValidationCase


def _test_roundtrip(source: BaseModel, target: BaseModel):
    """Test model equality for roundtrip tests"""

    assert type(target.array) is type(source.array)
    if isinstance(source.array, (np.ndarray, ZarrArray)):
        assert np.array_equal(target.array, np.array(source.array))
    elif isinstance(source.array, da.Array):
        if target.array.dtype == object and len(source.array.ravel()) > 0:
            # object equality doesn't really work well with dask
            # just check that the types match
            target_type = type(target.array.ravel()[0].compute())
            source_type = type(source.array.ravel()[0].compute())
            assert target_type is source_type
        else:
            assert np.all(da.equal(target.array, source.array))
    else:
        assert target.array == source.array

    assert target.array.dtype == source.array.dtype


def test_dunder_len(interface_cases, tmp_output_dir_func):
    """
    Each interface or proxy type should support __len__
    """
    case = ValidationCase(interface=interface_cases)
    if interface_cases.interface.name == "video":
        case.shape = (10, 10, 2, 3)
        case.dtype = np.uint8
        case.annotation_dtype = np.uint8
        case.annotation_shape = (10, 10, "*", 3)
    array = case.array(path=tmp_output_dir_func)
    instance = case.model(array=array)
    assert len(instance.array) == case.shape[0]


def test_interface_revalidate(all_passing_cases_instance):
    """
    An interface should revalidate with the output of its initial validation

    See: https://github.com/p2p-ld/numpydantic/pull/14
    """

    _ = type(all_passing_cases_instance)(array=all_passing_cases_instance.array)


@pytest.mark.json_schema
def test_interface_jsonschema(all_passing_cases_instance):
    """
    All interfaces should be able to generate json schema
    for all combinations of dtype and shape

    Note that this does not test for json schema correctness -
    see ndarray tests for that
    """
    _ = all_passing_cases_instance.model_json_schema()


@pytest.mark.xfail
def test_interface_rematch(interface_cases, tmp_output_dir_func):
    """
    All interfaces should match the results of the object they return after validation
    """
    array = interface_cases.make_array(path=tmp_output_dir_func)

    assert (
        Interface.match(interface_cases.interface.validate(array))
        is interface_cases.interface
    )


def test_interface_to_numpy_array(dtype_by_interface_instance):
    """
    All interfaces should be able to have the output of their validation stage
    coerced to a numpy array with np.array()
    """
    _ = np.array(dtype_by_interface_instance.array)


@pytest.mark.serialization
def test_interface_dump_json(dtype_by_interface_instance):
    """
    All interfaces should be able to dump to json
    """
    dtype_by_interface_instance.model_dump_json()


@pytest.mark.serialization
def test_interface_roundtrip_json(dtype_by_interface, tmp_output_dir_func):
    """
    All interfaces should be able to roundtrip to and from json
    """
    case = dtype_by_interface
    if "subclass" in case.id.lower():
        pytest.xfail()

    array = case.array(path=tmp_output_dir_func)
    instance = case.model(array=array)

    dumped_json = instance.model_dump_json(round_trip=True)
    model = instance.model_validate_json(dumped_json)
    _test_roundtrip(instance, model)


@pytest.mark.serialization
@pytest.mark.parametrize("case", ZERO_LENGTH_CASES_PASSING)
def test_roundtrip_json_zero_length(case, tmp_output_dir_func):
    """
    Arrays with zero-length dimensions should roundtrip to and from json,
    when supported by the array framework
    """
    if "zarr" in case.id.lower():
        pytest.skip("Zarr does not support reshaping arrays")

    array = case.array(path=tmp_output_dir_func)
    instance = case.model(array=array)

    dumped_json = instance.model_dump_json(round_trip=True)
    model = instance.model_validate_json(dumped_json)
    _test_roundtrip(instance, model)


@pytest.mark.serialization
@pytest.mark.parametrize("an_interface", Interface.interfaces())
def test_interface_mark_interface(an_interface):
    """
    All interfaces should be able to mark the current version and interface info
    """
    mark = an_interface.mark_interface()
    assert isinstance(mark, InterfaceMark)
    assert mark.name == an_interface.name
    assert mark.cls == an_interface.__name__
    assert mark.module == an_interface.__module__
    assert mark.version == version(mark.module.split(".")[0])


@pytest.mark.serialization
@pytest.mark.parametrize("valid", [True, False])
@pytest.mark.filterwarnings("ignore:Mismatch between serialized mark")
def test_interface_mark_roundtrip(dtype_by_interface, valid, tmp_output_dir_func):
    """
    All interfaces should be able to roundtrip with the marked interface,
    and a mismatch should raise a warning and attempt to proceed
    """
    if "subclass" in dtype_by_interface.id.lower():
        pytest.xfail()

    array = dtype_by_interface.array(path=tmp_output_dir_func)
    case = dtype_by_interface.model(array=array)

    dumped_json = case.model_dump_json(
        round_trip=True, context={"mark_interface": True}
    )

    data = json.loads(dumped_json)

    # ensure that we are a MarkedJson
    _ = MarkedJson.model_validate_json(json.dumps(data["array"]))

    if not valid:
        # ruin the version
        data["array"]["interface"]["version"] = "v99999999"
        dumped_json = json.dumps(data)

        with pytest.warns(match="Mismatch.*"):
            model = case.model_validate_json(dumped_json)
    else:
        model = case.model_validate_json(dumped_json)

    _test_roundtrip(case, model)



"""
Interface to numpy arrays
"""

from typing import Any, Literal, Union

from pydantic import BaseModel, SerializationInfo

from numpydantic.interface.interface import Interface, JsonDict

try:
    import numpy as np
    from numpy import ndarray

    ENABLED = True

except ImportError:  # pragma: no cover
    ENABLED = False
    ndarray = None
    np = None


class NumpyJsonDict(JsonDict):
    """
    JSON-able roundtrip representation of numpy array
    """

    type: Literal["numpy"]
    dtype: str
    value: list
    # allow shape to be None for backwards compatibility.
    shape: Union[tuple[int, ...], None] = None

    def to_array_input(self) -> ndarray:
        """
        Construct a numpy array
        """
        array = np.array(self.value, dtype=self.dtype)
        if self.shape is not None and array.shape != self.shape:
            array = self.reshape_input(array, self.shape)
        return array


class NumpyInterface(Interface):
    """
    Numpy :class:`~numpy.ndarray` s!
    """

    name = "numpy"
    input_types = (ndarray, list)
    return_type = ndarray
    json_model = NumpyJsonDict
    priority = -999
    """
    The numpy interface is usually the interface of last resort.
    We want to use any more specific interface that we might have,
    because the numpy interface checks for anything that could be coerced
    to a numpy array (see :meth:`.NumpyInterface.check` )
    """

    @classmethod
    def check(cls, array: Any) -> bool:
        """
        Check that this is in fact a numpy ndarray or something that can be
        coerced to one
        """
        if array is None:
            return False

        if isinstance(array, ndarray):
            return True
        elif isinstance(array, dict):
            return NumpyJsonDict.is_valid(array)
        else:
            try:
                _ = np.array(array)
                return True
            except Exception:
                return False

    def before_validation(self, array: Any) -> ndarray:
        """
        Coerce to an ndarray. We have already checked if coercion is possible
        in :meth:`.check`
        """
        if not isinstance(array, ndarray):
            array = np.array(array)

        try:
            if (
                issubclass(self.dtype, BaseModel)
                and len(array) > 0
                and isinstance(array.flat[0], dict)
            ):
                array = np.vectorize(lambda x: self.dtype(**x))(array)
        except TypeError:
            # fine, dtype isn't a type
            pass

        return array

    @classmethod
    def enabled(cls) -> bool:
        """Check that numpy is present in the environment"""
        return ENABLED

    @classmethod
    def to_json(
        cls, array: ndarray, info: SerializationInfo = None
    ) -> Union[list, JsonDict]:
        """
        Convert an array of :attr:`.return_type` to a JSON-compatible format using
        base python types
        """
        if not isinstance(array, np.ndarray):  # pragma: no cover
            array = np.array(array)

        json_array = [array.tolist()] if array.ndim == 0 else array.tolist()

        if info.round_trip:
            json_array = NumpyJsonDict(
                type=cls.name,
                dtype=str(array.dtype),
                value=json_array,
                shape=array.shape,
            )
        return json_array



"""
Types for numpydantic

Note that these are types as in python typing types, not classes.
"""

# ruff: noqa: D102

from typing import Any, Protocol, Tuple, Union, runtime_checkable

from numpydantic.vendor.nptyping import DType

ShapeType = Union[Tuple[int, ...], Any]
DtypeType = Union[str, type, Any, DType]


@runtime_checkable
class NDArrayType(Protocol):
    """A protocol for describing types that should be considered ndarrays"""

    @property
    def dtype(self) -> DtypeType: ...

    @property
    def shape(self) -> ShapeType: ...

    def __getitem__(self, key: Union[int, slice]) -> "NDArrayType": ...

    def __setitem__(self, key: Union[int, slice], value: "NDArrayType"): ...



"""
Test serialization-specific functionality that doesn't need to be
applied across every interface (use test_interface/test_interfaces for that
"""

import json
from pathlib import Path
from typing import Callable

import h5py
import numpy as np
import pytest

from numpydantic.serialization import _relativize_paths, _walk_and_apply, relative_path

pytestmark = pytest.mark.serialization


@pytest.fixture(scope="module")
def hdf5_at_path() -> Callable[[Path], None]:
    _path = ""

    def _hdf5_at_path(path: Path) -> None:
        nonlocal _path
        _path = path
        h5f = h5py.File(path, "w")
        _ = h5f.create_dataset("/data", data=np.array([[1, 2], [3, 4]]))
        _ = h5f.create_dataset("subpath/to/dataset", data=np.array([[1, 2], [4, 5]]))
        h5f.close()

    yield _hdf5_at_path

    Path(_path).unlink(missing_ok=True)


def test_relative_path(hdf5_at_path, tmp_output_dir, model_blank):
    """
    By default, we should make all paths relative to the cwd
    """
    out_path = tmp_output_dir / "relative.h5"
    hdf5_at_path(out_path)
    model = model_blank(array=(out_path, "/data"))
    rt = model.model_dump_json(round_trip=True)
    file = json.loads(rt)["array"]["file"]

    # should not be absolute
    assert not Path(file).is_absolute()
    # should be relative to cwd
    out_file = (Path.cwd() / file).resolve()
    assert out_file == out_path.resolve()


def test_relative_to_path(hdf5_at_path, tmp_output_dir, model_blank):
    """
    When explicitly passed a path to be ``relative_to`` ,
    relative to that instead of cwd
    """
    out_path = tmp_output_dir / "relative.h5"
    relative_to_path = Path(__file__) / "fake_dir" / "sub_fake_dir"
    expected_path = Path("../../../__tmp__/relative.h5")

    hdf5_at_path(out_path)
    model = model_blank(array=(out_path, "/data"))
    rt = model.model_dump_json(
        round_trip=True, context={"relative_to": str(relative_to_path)}
    )
    data = json.loads(rt)["array"]
    file = data["file"]

    # should not be absolute
    assert not Path(file).is_absolute()
    # should be expected path and reach the file
    assert Path(file) == expected_path
    assert (relative_to_path / file).resolve() == out_path.resolve()

    # we shouldn't have touched `/data` even though it is pathlike
    assert data["path"] == "/data"


def test_relative_to_root_dir():
    """
    The relativize function should ignore paths that are directories
    beneath the root directory (eg `/data`) even if they exist

    """
    # python 3.9 compat, which can't use negative indices
    test_path = [p for p in Path(__file__).resolve().parents][-2]

    test_data = {"some_field": str(test_path)}

    walked = _relativize_paths(test_data, relative_to=".")
    assert str(relative_path(test_path, Path(".").resolve())) != str(test_path)
    assert walked["some_field"] == str(test_path)


def test_absolute_path(hdf5_at_path, tmp_output_dir, model_blank):
    """
    When told, we make paths absolute
    """
    out_path = tmp_output_dir / "relative.h5"
    expected_dataset = "subpath/to/dataset"

    hdf5_at_path(out_path)
    model = model_blank(array=(out_path, expected_dataset))
    rt = model.model_dump_json(round_trip=True, context={"absolute_paths": True})
    data = json.loads(rt)["array"]
    file = data["file"]

    # should be absolute and equal to out_path
    assert Path(file).is_absolute()
    assert Path(file) == out_path.resolve()

    # shouldn't have absolutized subpath even if it's pathlike
    assert data["path"] == expected_dataset


def test_walk_and_apply():
    """
    Walk and apply should recursively apply a function to everything in a
    nesty structure
    """
    test = {
        "a": 1,
        "b": 1,
        "c": [
            {"a": 1, "b": {"a": 1, "b": 1}, "c": [1, 1, 1]},
            {"a": 1, "b": [1, 1, 1]},
        ],
    }

    def _mult_2(v, skip: bool = False):
        return v * 2

    def _assert_2(v, skip: bool = False):
        assert v == 2
        return v

    walked = _walk_and_apply(test, _mult_2)
    _walk_and_apply(walked, _assert_2)

    assert walked["a"] == 2
    assert walked["c"][0]["a"] == 2
    assert walked["c"][0]["b"]["a"] == 2
    assert all([w == 2 for w in walked["c"][0]["c"]])
    assert walked["c"][1]["a"] == 2
    assert all([w == 2 for w in walked["c"][1]["b"]])



"""
Interfaces between nptyping types and array backends
"""

from numpydantic.interface.dask import DaskInterface
from numpydantic.interface.hdf5 import H5ArrayPath, H5Interface
from numpydantic.interface.interface import (
    Interface,
    InterfaceMark,
    JsonDict,
    MarkedJson,
)
from numpydantic.interface.numpy import NumpyInterface
from numpydantic.interface.video import VideoInterface
from numpydantic.interface.zarr import ZarrArrayPath, ZarrInterface

__all__ = [
    "DaskInterface",
    "H5ArrayPath",
    "H5Interface",
    "Interface",
    "InterfaceMark",
    "JsonDict",
    "MarkedJson",
    "NumpyInterface",
    "VideoInterface",
    "ZarrArrayPath",
    "ZarrInterface",
]



import sys
from typing import Union

import numpy as np
from pydantic import BaseModel

from numpydantic.dtype import Float, Integer, Number
from numpydantic.testing.helpers import ValidationCase, merged_product
from numpydantic.testing.interfaces import (
    DaskCase,
    HDF5Case,
    HDF5CompoundCase,
    NumpyCase,
    VideoCase,
    ZarrCase,
    ZarrDirCase,
    ZarrNestedCase,
    ZarrZipCase,
)

if sys.version_info.minor >= 10:
    from typing import TypeAlias

    YES_PIPE = True
else:
    from typing_extensions import TypeAlias

    YES_PIPE = False


class BasicModel(BaseModel):
    x: int


class BadModel(BaseModel):
    x: int


class SubClass(BasicModel):
    pass


# --------------------------------------------------
# Annotations
# --------------------------------------------------

RGB_UNION = (("*", "*"), ("*", "*", 3), ("*", "*", 3, 4))
UNION_TYPE: TypeAlias = Union[np.uint32, np.float32]

SHAPE_CASES = (
    ValidationCase(shape=(10, 10, 2, 2), passes=True, id="valid shape"),
    ValidationCase(shape=(10, 10, 2), passes=False, id="missing dimension"),
    ValidationCase(shape=(10, 10, 2, 2, 2), passes=False, id="extra dimension"),
    ValidationCase(shape=(11, 10, 2, 2), passes=False, id="dimension too large"),
    ValidationCase(shape=(9, 10, 2, 2), passes=False, id="dimension too small"),
    ValidationCase(shape=(10, 10, 1, 1), passes=True, id="wildcard smaller"),
    ValidationCase(shape=(10, 10, 3, 3), passes=True, id="wildcard larger"),
    ValidationCase(
        annotation_shape=RGB_UNION, shape=(5, 5), passes=True, id="Union 2D"
    ),
    ValidationCase(
        annotation_shape=RGB_UNION, shape=(5, 5, 3), passes=True, id="Union 3D"
    ),
    ValidationCase(
        annotation_shape=RGB_UNION, shape=(5, 5, 3, 4), passes=True, id="Union 4D"
    ),
    ValidationCase(
        annotation_shape=RGB_UNION,
        shape=(5, 5, 4),
        passes=False,
        id="Union incorrect 3D",
    ),
    ValidationCase(
        annotation_shape=RGB_UNION,
        shape=(5, 5, 3, 6),
        passes=False,
        id="Union incorrect 4D",
    ),
    ValidationCase(
        annotation_shape=RGB_UNION,
        shape=(5, 5, 4, 6),
        passes=False,
        id="Union incorrect both",
    ),
    ValidationCase(
        annotation_shape=("*", "..."),
        shape=tuple(),
        id="scalar",
        marks={"scalar"},
        passes=True,
    ),
    ValidationCase(
        annotation_shape=("*", "*"),
        shape=tuple(),
        id="scalar-min-dimensions",
        marks={"scalar"},
        passes=False,
    ),
    ValidationCase(
        annotation_shape=("*", 3), shape=(0, 3), id="zero-length", passes=True
    ),
)
"""
Base Shape cases
"""


DTYPE_CASES = [
    ValidationCase(dtype=float, passes=True, id="float"),
    ValidationCase(dtype=int, passes=False, id="int"),
    ValidationCase(dtype=np.uint8, passes=False, id="uint8"),
    ValidationCase(annotation_dtype=Number, dtype=int, passes=True, id="number-int"),
    ValidationCase(
        annotation_dtype=Number, dtype=float, passes=True, id="number-float"
    ),
    ValidationCase(
        annotation_dtype=Number, dtype=np.uint8, passes=True, id="number-uint8"
    ),
    ValidationCase(
        annotation_dtype=Number, dtype=np.float16, passes=True, id="number-float16"
    ),
    ValidationCase(annotation_dtype=Number, dtype=str, passes=False, id="number-str"),
    ValidationCase(annotation_dtype=Integer, dtype=int, passes=True, id="integer-int"),
    ValidationCase(
        annotation_dtype=Integer, dtype=np.uint8, passes=True, id="integer-uint8"
    ),
    ValidationCase(
        annotation_dtype=Integer, dtype=float, passes=False, id="integer-float"
    ),
    ValidationCase(
        annotation_dtype=Integer, dtype=np.float32, passes=False, id="integer-float32"
    ),
    ValidationCase(annotation_dtype=Integer, dtype=str, passes=False, id="integer-str"),
    ValidationCase(annotation_dtype=Float, dtype=float, passes=True, id="float-float"),
    ValidationCase(
        annotation_dtype=Float, dtype=np.float32, passes=True, id="float-float32"
    ),
    ValidationCase(annotation_dtype=Float, dtype=int, passes=False, id="float-int"),
    ValidationCase(
        annotation_dtype=Float, dtype=np.uint8, passes=False, id="float-uint8"
    ),
    ValidationCase(annotation_dtype=Float, dtype=str, passes=False, id="float-str"),
    ValidationCase(annotation_dtype=str, dtype=str, passes=True, id="str-str"),
    ValidationCase(annotation_dtype=str, dtype=int, passes=False, id="str-int"),
    ValidationCase(annotation_dtype=str, dtype=float, passes=False, id="str-float"),
    ValidationCase(
        annotation_dtype=np.str_,
        dtype=str,
        passes=True,
        id="np_str-str",
        marks={"np_str", "str"},
    ),
    ValidationCase(
        annotation_dtype=np.str_,
        dtype=np.str_,
        passes=True,
        id="np_str-np_str",
        marks={"np_str", "str"},
    ),
    ValidationCase(
        annotation_dtype=(int, np.str_),
        dtype=str,
        passes=True,
        id="tuple_np_str-str",
        marks={"np_str", "str", "tuple"},
    ),
    ValidationCase(
        annotation_dtype=BasicModel,
        dtype=BasicModel,
        passes=True,
        id="model-model",
        marks={"model"},
    ),
    ValidationCase(
        annotation_dtype=BasicModel,
        dtype=BadModel,
        passes=False,
        id="model-badmodel",
        marks={"model"},
    ),
    ValidationCase(
        annotation_dtype=BasicModel,
        dtype=int,
        passes=False,
        id="model-int",
        marks={"model"},
    ),
    ValidationCase(
        annotation_dtype=BasicModel,
        dtype=SubClass,
        passes=True,
        id="model-subclass",
        marks={"model"},
    ),
    ValidationCase(
        annotation_dtype=UNION_TYPE,
        dtype=np.uint32,
        passes=True,
        id="union-type-uint32",
        marks={"union"},
    ),
    ValidationCase(
        annotation_dtype=UNION_TYPE,
        dtype=np.float32,
        passes=True,
        id="union-type-float32",
        marks={"union"},
    ),
    ValidationCase(
        annotation_dtype=UNION_TYPE,
        dtype=np.uint64,
        passes=False,
        id="union-type-uint64",
        marks={"union"},
    ),
    ValidationCase(
        annotation_dtype=UNION_TYPE,
        dtype=np.float64,
        passes=False,
        id="union-type-float64",
        marks={"union"},
    ),
    ValidationCase(
        annotation_dtype=UNION_TYPE,
        dtype=str,
        passes=False,
        id="union-type-str",
        marks={"union"},
    ),
]
"""
Base Dtype cases
"""


if YES_PIPE:
    UNION_PIPE: TypeAlias = np.uint32 | np.float32

    DTYPE_CASES.extend(
        [
            ValidationCase(
                annotation_dtype=UNION_PIPE,
                dtype=np.uint32,
                passes=True,
                id="union-pipe-uint32",
                marks={"union", "pipe_union"},
            ),
            ValidationCase(
                annotation_dtype=UNION_PIPE,
                dtype=np.float32,
                passes=True,
                id="union-pipe-float32",
                marks={"union", "pipe_union"},
            ),
            ValidationCase(
                annotation_dtype=UNION_PIPE,
                dtype=np.uint64,
                passes=False,
                id="union-pipe-uint64",
                marks={"union", "pipe_union"},
            ),
            ValidationCase(
                annotation_dtype=UNION_PIPE,
                dtype=np.float64,
                passes=False,
                id="union-pipe-float64",
                marks={"union", "pipe_union"},
            ),
            ValidationCase(
                annotation_dtype=UNION_PIPE,
                dtype=str,
                passes=False,
                id="union-pipe-str",
                marks={"union", "pipe_union"},
            ),
        ]
    )

INTERFACE_CASES = [
    ValidationCase(interface=NumpyCase, id="numpy"),
    ValidationCase(interface=HDF5Case, id="hdf5"),
    ValidationCase(interface=HDF5CompoundCase, id="hdf5_compound"),
    ValidationCase(interface=DaskCase, id="dask"),
    ValidationCase(interface=ZarrCase, id="zarr"),
    ValidationCase(interface=ZarrDirCase, id="zarr_dir"),
    ValidationCase(interface=ZarrZipCase, id="zarr_zip"),
    ValidationCase(interface=ZarrNestedCase, id="zarr_nested"),
    ValidationCase(interface=VideoCase, id="video"),
]
"""
All the interface cases
"""


DTYPE_AND_SHAPE_CASES = merged_product(SHAPE_CASES, DTYPE_CASES)
"""
Merged product of dtype and shape cases
"""
DTYPE_AND_SHAPE_CASES_PASSING = merged_product(
    SHAPE_CASES, DTYPE_CASES, conditions={"passes": True}
)
"""
Merged product of dtype and shape cases that are valid
"""

DTYPE_AND_INTERFACE_CASES = merged_product(INTERFACE_CASES, DTYPE_CASES)
"""
Merged product of dtype and interface cases
"""
DTYPE_AND_INTERFACE_CASES_PASSING = merged_product(
    INTERFACE_CASES, DTYPE_CASES, conditions={"passes": True}
)
"""
Merged product of dtype and interface cases that pass
"""

ALL_CASES = merged_product(SHAPE_CASES, DTYPE_CASES, INTERFACE_CASES)
"""
Merged product of all cases - dtype, shape, and interface
"""
ALL_CASES_PASSING = merged_product(
    SHAPE_CASES, DTYPE_CASES, INTERFACE_CASES, conditions={"passes": True}
)
"""
Merged product of all cases, but only those that pass
"""

ZERO_LENGTH_CASES_PASSING = [c for c in ALL_CASES_PASSING if "zero-length" in c.id]



import json
from typing import Any, Optional, Union

import numpy as np
import pytest
from pydantic import (
    BaseModel,
    Field,
    ValidationError,
)

from numpydantic import NDArray, Shape, dtype
from numpydantic.dtype import Number
from numpydantic.exceptions import DtypeError


@pytest.mark.json_schema
def test_ndarray_type():
    class Model(BaseModel):
        array: NDArray[Shape["2 x, * y"], Number]
        array_any: Optional[NDArray[Any, Any]] = None

    schema = Model.model_json_schema()
    assert schema["properties"]["array"]["items"] == {
        "items": {"type": "number"},
        "type": "array",
    }
    assert schema["properties"]["array"]["maxItems"] == 2
    assert schema["properties"]["array"]["minItems"] == 2

    # models should instantiate correctly!
    _ = Model(array=np.zeros((2, 3)))

    with pytest.raises(ValidationError):
        _ = Model(array=np.zeros((4, 6)))

    with pytest.raises(ValidationError):
        _ = Model(array=np.ones((2, 3), dtype=bool))

    _ = Model(array=np.zeros((2, 3)), array_any=np.ones((3, 4, 5)))


@pytest.mark.dtype
@pytest.mark.json_schema
def test_schema_unsupported_type():
    """
    Complex numbers should just be made with an `any` schema
    """

    class Model(BaseModel):
        array: NDArray[Shape["2 x, * y"], complex]

    schema = Model.model_json_schema()
    assert schema["properties"]["array"]["items"] == {
        "items": {},
        "type": "array",
    }


@pytest.mark.dtype
@pytest.mark.json_schema
def test_schema_tuple():
    """
    Types specified as tuples should have their schemas as a union
    """

    class Model(BaseModel):
        array: NDArray[Shape["2 x, * y"], (np.uint8, np.uint16)]

    schema = Model.model_json_schema()
    assert "anyOf" in schema["properties"]["array"]["items"]["items"]
    conditions = schema["properties"]["array"]["items"]["items"]["anyOf"]

    assert all([i["type"] == "integer" for i in conditions])
    assert sorted([i["maximum"] for i in conditions]) == [255, 65535]
    assert all([i["minimum"] == 0 for i in conditions])


@pytest.mark.dtype
@pytest.mark.json_schema
def test_schema_number():
    """
    np.numeric should just be the float schema
    """

    class Model(BaseModel):
        array: NDArray[Shape["2 x, * y"], np.number]

    schema = Model.model_json_schema()
    assert schema["properties"]["array"]["items"] == {
        "items": {"type": "number"},
        "type": "array",
    }


def test_ndarray_union():
    generator = np.random.default_rng()

    class Model(BaseModel):
        array: Optional[
            Union[
                NDArray[Shape["* x, * y"], Number],
                NDArray[Shape["* x, * y, 3 r_g_b"], Number],
                NDArray[Shape["* x, * y, 3 r_g_b, 4 r_g_b_a"], Number],
            ]
        ] = Field(None)

    _ = Model()
    _ = Model(array=generator.random((5, 10)))
    _ = Model(array=generator.random((5, 10, 3)))
    _ = Model(array=generator.random((5, 10, 3, 4)))

    with pytest.raises(ValidationError):
        _ = Model(array=generator.random((5,)))

    with pytest.raises(ValidationError):
        _ = Model(array=generator.random((5, 10, 4)))

    with pytest.raises(ValidationError):
        _ = Model(array=generator.random((5, 10, 3, 6)))

    with pytest.raises(ValidationError):
        _ = Model(array=generator.random((5, 10, 4, 6)))


@pytest.mark.shape
@pytest.mark.dtype
@pytest.mark.parametrize("dtype", dtype.Number)
def test_ndarray_unparameterized(dtype):
    """
    NDArray without any parameters is any shape, any type
    """
    generator = np.random.default_rng()

    class Model(BaseModel):
        array: NDArray

    # not very sophisticated fuzzing of "any shape"
    test_cases = 10
    for _ in range(test_cases):
        n_dimensions = generator.integers(1, 8)
        dim_sizes = generator.integers(1, 7, size=n_dimensions)
        _ = Model(array=np.zeros(dim_sizes, dtype=dtype))


@pytest.mark.shape
def test_ndarray_any():
    """
    using :class:`typing.Any` in for the shape means any shape
    """
    generator = np.random.default_rng()

    class Model(BaseModel):
        array: NDArray[Any, np.uint8]

    # not very sophisticated fuzzing of "any shape"
    test_cases = 100
    for _ in range(test_cases):
        n_dimensions = generator.integers(1, 8)
        dim_sizes = generator.integers(1, 16, size=n_dimensions)
        _ = Model(array=np.zeros(dim_sizes, dtype=np.uint8))


def test_ndarray_coercion():
    """
    Coerce lists to arrays
    """

    class Model(BaseModel):
        array: NDArray[Shape["* x"], Number]

    amod = Model(array=[1, 2, 3, 4.5])
    assert np.allclose(amod.array, np.array([1, 2, 3, 4.5]))
    with pytest.raises(ValidationError):
        amod = Model(array=["a", "b", "c"])


@pytest.mark.shape
def test_shape_ellipsis():
    """
    Test that ellipsis is a wildcard, rather than "repeat the last index"
    """

    class MyModel(BaseModel):
        array: NDArray[Shape["1, 2, ..."], Number]

    _ = MyModel(array=np.zeros((1, 2, 3, 4, 5)))


@pytest.mark.serialization
def test_ndarray_serialize():
    """
    Arrays should be dumped to a list when using json, but kept as ndarray otherwise
    """

    class Model(BaseModel):
        array: NDArray[Any, Number]

    mod = Model(array=np.random.default_rng().random((3, 3)))
    mod_str = mod.model_dump_json()
    mod_json = json.loads(mod_str)
    assert isinstance(mod_json["array"], list)

    # but when we just dump to a dict we don't coerce
    mod_dict = mod.model_dump()
    assert isinstance(mod_dict["array"], np.ndarray)


_json_schema_types = [
    *[(t, float) for t in dtype.Float],
    *[(t, int) for t in dtype.Integer],
]


@pytest.mark.json_schema
def test_json_schema_basic(array_model):
    """
    NDArray types should correctly generate a list of lists JSON schema
    """
    shape = (15, 10)
    dtype = float
    model = array_model(shape, dtype)
    schema = model.model_json_schema()
    field = schema["properties"]["array"]

    # outer shape
    assert field["maxItems"] == shape[0]
    assert field["minItems"] == shape[0]
    assert field["type"] == "array"

    # inner shape
    inner = field["items"]
    assert inner["minItems"] == shape[1]
    assert inner["maxItems"] == shape[1]
    assert inner["items"]["type"] == "number"


@pytest.mark.dtype
@pytest.mark.json_schema
@pytest.mark.parametrize("dtype", [*dtype.Integer, *dtype.Float])
def test_json_schema_dtype_single(dtype, array_model):
    """
    dtypes should have correct mins and maxes set, and store the source dtype
    """
    if issubclass(dtype, np.floating):
        info = np.finfo(dtype)
        min_val = info.min
        max_val = info.max
        schema_type = "number"
    elif issubclass(dtype, np.integer):
        info = np.iinfo(dtype)
        min_val = info.min
        max_val = info.max
        schema_type = "integer"
    else:
        raise ValueError("These should all be numpy types!")

    shape = (15, 10)
    model = array_model(shape, dtype)
    schema = model.model_json_schema()
    inner_type = schema["properties"]["array"]["items"]["items"]
    assert inner_type["minimum"] == min_val
    assert inner_type["maximum"] == max_val
    assert inner_type["type"] == schema_type
    assert schema["properties"]["array"]["dtype"] == ".".join(
        [dtype.__module__, dtype.__name__]
    )


@pytest.mark.dtype
@pytest.mark.json_schema
@pytest.mark.parametrize(
    "dtype,expected",
    [
        (dtype.Integer, "integer"),
        (dtype.Float, "number"),
        (dtype.Bool, "boolean"),
        (int, "integer"),
        (float, "number"),
        (bool, "boolean"),
        (complex, "any"),
    ],
)
def test_json_schema_dtype_builtin(dtype, expected, array_model):
    """
    Using builtin or generic (eg. `dtype.Integer` ) dtypes should
    make a simple json schema without mins/maxes/dtypes.
    """
    model = array_model(dtype=dtype)
    schema = model.model_json_schema()
    inner_type = schema["properties"]["array"]["items"]["items"]
    if expected == "any":
        assert inner_type == {}
    else:
        assert inner_type["type"] == expected


@pytest.mark.dtype
@pytest.mark.json_schema
def test_json_schema_dtype_model():
    """
    Pydantic models can be used in arrays as dtypes
    """

    class TestModel(BaseModel):
        x: int
        y: int
        z: int

    class MyModel(BaseModel):
        array: NDArray[Shape["*, *"], TestModel]

    schema = MyModel.model_json_schema()
    # we should have a "$defs" with TestModel in it,
    # and our array should be objects of that type
    assert schema["properties"]["array"]["items"]["items"] == {
        "$ref": "#/$defs/TestModel"
    }
    # we don't test pydantic' generic json schema model generation,
    # just that one was defined
    assert "TestModel" in schema["$defs"]


def _recursive_array(schema):
    assert "$defs" in schema
    # get the key uses for the array
    array_key = list(schema["$defs"].keys())[0]

    # the array property should be a ref to the recursive array
    # get the innermost part of the field schema
    field_schema = schema["properties"]["array"]
    while "items" in field_schema:
        field_schema = field_schema["items"]
    assert field_schema["$ref"] == f"#/$defs/{array_key}"

    # and the recursive array should indeed be recursive...
    # specifically it should be an array whose items can be itself or
    # of the type specified by the dtype
    any_of = schema["$defs"][array_key]["anyOf"]
    assert any_of[0]["items"]["$ref"] == f"#/$defs/{array_key}"
    assert any_of[0]["type"] == "array"
    # here we are just assuming that it's a uint8 array..
    assert any_of[1]["type"] == "integer"
    assert any_of[1]["maximum"] == 255
    assert any_of[1]["minimum"] == 0


@pytest.mark.shape
@pytest.mark.json_schema
def test_json_schema_ellipsis():
    """
    NDArray types should create a recursive JSON schema for any-shaped arrays
    """

    class AnyShape(BaseModel):
        array: NDArray[Shape["*, ..."], np.uint8]

    schema = AnyShape.model_json_schema()
    _recursive_array(schema)

    class ConstrainedAnyShape(BaseModel):
        array: NDArray[Shape["3, 4, ..."], np.uint8]

    schema = ConstrainedAnyShape.model_json_schema()
    _recursive_array(schema)


def test_instancecheck():
    """
    NDArray should handle ``isinstance()`` s.t. valid arrays are ``True``
    and invalid arrays are ``False``

    We don't make this test exhaustive because correctness of validation
    is tested elsewhere. We are just testing that the type checking works
    """
    array_type = NDArray[Shape["1, 2, 3"], int]

    assert isinstance(np.zeros((1, 2, 3), dtype=int), array_type)
    assert not isinstance(np.zeros((2, 2, 3), dtype=int), array_type)
    assert not isinstance(np.zeros((1, 2, 3), dtype=float), array_type)

    def my_function(array: NDArray[Shape["1, 2, 3"], int]):
        return array

    my_function(np.zeros((1, 2, 3), int))


def test_callable():
    """
    NDArray objects are callable to validate and cast
    Don't test validation here, just that we can be called
    """
    annotation = NDArray[Shape["3"], int]
    array = np.array([1, 2, 3], dtype=int)
    validated = annotation(array)
    assert validated is array

    with pytest.raises(DtypeError):
        _ = annotation(np.zeros((1, 2, 3)))



"""
Interface for Dask arrays
"""

from typing import Any, Iterable, List, Literal, Optional, Union

import numpy as np
from pydantic import BaseModel, SerializationInfo

from numpydantic.interface.interface import Interface, JsonDict
from numpydantic.types import DtypeType, NDArrayType

try:
    from dask.array import from_array
    from dask.array.core import Array as DaskArray
except ImportError:  # pragma: no cover
    DaskArray = None


def _as_tuple(a_list: Any) -> tuple:
    """Make a list of list into a tuple of tuples"""
    return tuple(
        [_as_tuple(item) if isinstance(item, list) else item for item in a_list]
    )


class DaskJsonDict(JsonDict):
    """
    Round-trip json serialized form of a dask array
    """

    type: Literal["dask"]
    name: str
    chunks: Iterable[tuple[int, ...]]
    dtype: str
    shape: Union[tuple[int, ...], None] = None
    value: list

    def to_array_input(self) -> DaskArray:
        """Construct a dask array"""
        np_array = np.array(self.value, dtype=self.dtype)
        if self.shape is not None and np_array.shape != self.shape:
            np_array = self.reshape_input(np_array, self.shape)
        array = from_array(
            np_array,
            name=self.name,
            chunks=_as_tuple(self.chunks),
        )
        return array


class DaskInterface(Interface):
    """
    Interface for Dask :class:`~dask.array.core.Array`
    """

    name = "dask"
    input_types = (DaskArray, dict)
    return_type = DaskArray
    json_model = DaskJsonDict

    @classmethod
    def check(cls, array: Any) -> bool:
        """
        check if array is a dask array
        """
        if DaskArray is None:  # pragma: no cover - no tests for interface deps atm
            return False
        elif isinstance(array, DaskArray):
            return True
        elif isinstance(array, dict):
            return DaskJsonDict.is_valid(array)
        else:
            return False

    def before_validation(self, array: DaskArray) -> NDArrayType:
        """
        Try and coerce dicts that should be model objects into the model objects
        """
        try:
            if issubclass(self.dtype, BaseModel):
                flat_array = array.reshape(-1)
                if len(flat_array) == 0:
                    return array

                if isinstance(flat_array[0].compute(), dict):

                    def _chunked_to_model(array: np.ndarray) -> np.ndarray:
                        def _vectorized_to_model(
                            item: Union[dict, BaseModel],
                        ) -> BaseModel:
                            if not isinstance(item, self.dtype):
                                return self.dtype(**item)
                            else:  # pragma: no cover
                                return item

                        return np.vectorize(_vectorized_to_model)(array)

                    array = array.map_blocks(_chunked_to_model, dtype=self.dtype)
        except TypeError:
            # fine, dtype isn't a type
            pass
        return array

    def get_object_dtype(self, array: NDArrayType) -> DtypeType:
        """Dask arrays require a compute() call to retrieve a single value"""
        flat_array = array.reshape(-1)
        if len(flat_array) == 0:
            return Any
        else:
            return type(flat_array[0].compute())

    @classmethod
    def enabled(cls) -> bool:
        """check if we successfully imported dask"""
        return DaskArray is not None

    @classmethod
    def to_json(
        cls, array: DaskArray, info: Optional[SerializationInfo] = None
    ) -> Union[List, DaskJsonDict]:
        """
        Convert an array to a JSON serializable array by first converting to a numpy
        array and then to a list.

        .. note::

            This is likely a very memory intensive operation if you are using dask for
            large arrays. This can't be avoided, since the creation of the json string
            happens in-memory with Pydantic, so you are likely looking for a different
            method of serialization here using the python object itself rather than
            its JSON representation.
        """
        np_array = np.array(array)
        as_json = np_array.tolist()
        if not isinstance(as_json, list):
            as_json = [as_json]
        if info.round_trip:
            as_json = DaskJsonDict(
                type=cls.name,
                value=as_json,
                name=array.name,
                chunks=array.chunks,
                dtype=str(np_array.dtype),
                shape=array.shape,
            )
        return as_json



import pytest

from numpydantic.testing.cases import (
    ALL_CASES,
    ALL_CASES_PASSING,
    DTYPE_AND_INTERFACE_CASES_PASSING,
)
from numpydantic.testing.helpers import InterfaceCase, ValidationCase, merge_cases
from numpydantic.testing.interfaces import (
    DaskCase,
    HDF5Case,
    NumpyCase,
    VideoCase,
    ZarrCase,
    ZarrDirCase,
    ZarrNestedCase,
)


@pytest.fixture(
    scope="function",
    params=[
        pytest.param(
            NumpyCase,
            marks=pytest.mark.numpy,
            id="numpy",
        ),
        pytest.param(
            HDF5Case,
            marks=pytest.mark.hdf5,
            id="h5-array-path",
        ),
        pytest.param(
            DaskCase,
            marks=pytest.mark.dask,
            id="dask",
        ),
        pytest.param(
            ZarrCase,
            marks=pytest.mark.zarr,
            id="zarr-memory",
        ),
        pytest.param(
            ZarrNestedCase,
            marks=pytest.mark.zarr,
            id="zarr-nested",
        ),
        pytest.param(
            ZarrDirCase,
            marks=pytest.mark.zarr,
            id="zarr-dir",
        ),
        pytest.param(VideoCase, marks=pytest.mark.video, id="video"),
    ],
)
def interface_cases(request) -> InterfaceCase:
    """
    Fixture for combinatoric tests across all interface cases
    """
    return request.param


@pytest.fixture(
    params=(pytest.param(p, id=p.id, marks=p.pytest_marks) for p in ALL_CASES)
)
def all_cases(interface_cases, request) -> ValidationCase:
    """
    Combinatoric testing for all dtype, shape, and interface cases.

    This is a very expensive fixture! Only use it for core functionality
    that we want to be sure is *very true* in every circumstance,
    INCLUDING invalid combinations of annotations and arrays.
    Typically, that means only use this in `test_interfaces.py`
    """

    case = merge_cases(request.param, ValidationCase(interface=interface_cases))
    if case.skip():
        pytest.skip()
    return case


@pytest.fixture(
    params=(pytest.param(p, id=p.id, marks=p.pytest_marks) for p in ALL_CASES_PASSING)
)
def all_passing_cases(request) -> ValidationCase:
    """
    Combinatoric testing for all dtype, shape, and interface cases,
    but only the combinations that we expect to pass.

    This is a very expensive fixture! Only use it for core functionality
    that we want to be sure is *very true* in every circumstance.
    Typically, that means only use this in `test_interfaces.py`
    """
    return request.param


@pytest.fixture()
def all_cases_instance(all_cases, tmp_output_dir_func):
    """
    all_cases but with an instantiated model
    Args:
        all_cases:

    Returns:

    """
    array = all_cases.array(path=tmp_output_dir_func)
    instance = all_cases.model(array=array)
    return instance


@pytest.fixture()
def all_passing_cases_instance(all_passing_cases, tmp_output_dir_func):
    """
    all_cases but with an instantiated model
    Args:
        all_cases:

    Returns:

    """
    array = all_passing_cases.array(path=tmp_output_dir_func)
    instance = all_passing_cases.model(array=array)
    return instance


@pytest.fixture(
    params=(
        pytest.param(p, id=p.id, marks=p.pytest_marks)
        for p in DTYPE_AND_INTERFACE_CASES_PASSING
    )
)
def dtype_by_interface(request):
    """
    Tests for all dtypes by all interfaces
    """
    return request.param


@pytest.fixture()
def dtype_by_interface_instance(dtype_by_interface, tmp_output_dir_func):
    array = dtype_by_interface.array(path=tmp_output_dir_func)
    instance = dtype_by_interface.model(array=array)
    return instance



"""
Tests for the interface base model,
for tests that should apply to all interfaces, use ``test_interfaces.py``
"""

import gc
from typing import Literal

import numpy as np
import pytest
from pydantic import ValidationError

from numpydantic.interface import (
    Interface,
    InterfaceMark,
    JsonDict,
    MarkedJson,
    NumpyInterface,
)
from numpydantic.interface.interface import V


class MyJsonDict(JsonDict):
    type: Literal["my_json_dict"]
    field: str
    number: int

    def to_array_input(self) -> V:
        dumped = self.model_dump()
        dumped["extra_input_param"] = True
        return dumped


@pytest.fixture(scope="module")
def interfaces():
    """Define test interfaces in this module, and delete afterwards"""
    interfaces_enabled = True

    class Interface1(Interface):
        input_types = (list,)
        return_type = tuple
        priority = 1000
        checked = False

        @classmethod
        def check(cls, array):
            cls.checked = True
            return isinstance(array, list)

        @classmethod
        def enabled(cls) -> bool:
            return interfaces_enabled

    Interface2 = type("Interface2", Interface1.__bases__, dict(Interface1.__dict__))
    Interface2.checked = False
    Interface2.priority = 999

    class Interface3(Interface1):
        priority = 998
        checked = False

        @classmethod
        def enabled(cls) -> bool:
            return False

    class Interface4(Interface3):
        priority = 997
        checked = False

        @classmethod
        def enabled(cls) -> bool:
            return interfaces_enabled

    class Interfaces:
        interface1 = Interface1
        interface2 = Interface2
        interface3 = Interface3
        interface4 = Interface4

    yield Interfaces
    # Interface.__subclasses__().remove(Interface1)
    # Interface.__subclasses__().remove(Interface2)
    del Interfaces
    del Interface1
    del Interface2
    del Interface3
    del Interface4
    interfaces_enabled = False
    gc.collect()


def test_interface_match_error(interfaces):
    """
    Test that `match` and `match_output` raises errors when no or multiple matches
    are found
    """
    with pytest.raises(ValueError) as e:
        Interface.match([1, 2, 3])
    assert "Interface1" in str(e.value)
    assert "Interface2" in str(e.value)

    with pytest.raises(ValueError) as e:
        Interface.match(([1, 2, 3], ["hey"]))
    assert "No matching interfaces" in str(e.value)

    with pytest.raises(ValueError) as e:
        Interface.match_output((1, 2, 3))
    assert "Interface1" in str(e.value)
    assert "Interface2" in str(e.value)

    with pytest.raises(ValueError) as e:
        Interface.match_output("hey")
    assert "No matching interfaces" in str(e.value)


def test_interface_match_fast(interfaces):
    """
    fast matching should return as soon as an interface is found
    and not raise an error for duplicates
    """
    Interface.interfaces()[0].checked = False
    Interface.interfaces()[1].checked = False
    # this doesnt' raise an error
    matched = Interface.match([1, 2, 3], fast=True)
    assert matched == Interface.interfaces()[0]
    assert Interface.interfaces()[0].checked
    assert not Interface.interfaces()[1].checked


def test_interface_enabled(interfaces):
    """
    An interface shouldn't be included if it's not enabled
    """
    assert not interfaces.interface3.enabled()
    assert interfaces.interface3 not in Interface.interfaces()


def test_interface_type_lists():
    """
    Seems like a silly test, but ensure that our return types and input types
    lists have all the class attrs
    """
    for interface in Interface.interfaces():

        if isinstance(interface.input_types, (list, tuple)):
            for atype in interface.input_types:
                assert atype in Interface.input_types()
        else:
            assert interface.input_types in Interface.input_types()

        if isinstance(interface.return_type, (list, tuple)):
            for atype in interface.return_type:
                assert atype in Interface.return_types()
        else:
            assert interface.return_type in Interface.return_types()


def test_interfaces_sorting():
    """
    Interfaces should be returned in descending order of priority
    """
    ifaces = Interface.interfaces()
    priorities = [i.priority for i in ifaces]
    assert (np.diff(priorities) <= 0).all()


def test_interface_with_disabled(interfaces):
    """
    Get all interfaces, even if not enabled
    """
    ifaces = Interface.interfaces(with_disabled=True)
    assert interfaces.interface3 in ifaces


def test_interface_recursive(interfaces):
    """
    Get all interfaces, including subclasses of subclasses
    """
    ifaces = Interface.interfaces()
    assert issubclass(interfaces.interface4, interfaces.interface3)
    assert issubclass(interfaces.interface3, interfaces.interface1)
    assert issubclass(interfaces.interface1, Interface)
    assert interfaces.interface4 in ifaces


@pytest.mark.serialization
def test_jsondict_is_valid():
    """
    A JsonDict should return a bool true/false if it is valid or not,
    and raise an error when requested
    """
    invalid = {"doesnt": "have", "the": "props"}
    valid = {"type": "my_json_dict", "field": "a_field", "number": 1}
    assert MyJsonDict.is_valid(valid)
    assert not MyJsonDict.is_valid(invalid)
    with pytest.raises(ValidationError):
        assert not MyJsonDict.is_valid(invalid, raise_on_error=True)


@pytest.mark.serialization
def test_jsondict_handle_input():
    """
    JsonDict should be able to parse a valid dict and return it to the input format
    """
    valid = {"type": "my_json_dict", "field": "a_field", "number": 1}
    instantiated = MyJsonDict(**valid)
    expected = {
        "type": "my_json_dict",
        "field": "a_field",
        "number": 1,
        "extra_input_param": True,
    }

    for item in (valid, instantiated):
        result = MyJsonDict.handle_input(item)
        assert result == expected


@pytest.mark.serialization
@pytest.mark.parametrize("interface", Interface.interfaces())
def test_interface_mark_match_by_name(interface):
    """
    Interface mark should match an interface by its name
    """
    # other parts don't matter
    mark = InterfaceMark(module="fake", cls="fake", version="fake", name=interface.name)
    fake_mark = InterfaceMark(
        module="fake", cls="fake", version="fake", name="also_fake"
    )
    assert mark.match_by_name() is interface
    assert fake_mark.match_by_name() is None


@pytest.mark.serialization
def test_marked_json_try_cast():
    """
    MarkedJson.try_cast should try and cast to a markedjson!
    returning the value unchanged if it's not a match
    """
    valid = {"interface": NumpyInterface.mark_interface(), "value": [[1, 2], [3, 4]]}
    invalid = [1, 2, 3, 4, 5]
    mimic = {"interface": "not really", "value": "still not really"}

    assert isinstance(MarkedJson.try_cast(valid), MarkedJson)
    assert MarkedJson.try_cast(invalid) is invalid
    assert MarkedJson.try_cast(mimic) is mimic



"""
Metaprogramming functions for numpydantic to modify itself :)
"""

from pathlib import Path
from warnings import warn

from numpydantic.interface import Interface

_BUILTIN_IMPORTS = ("import typing", "import pathlib")
_PREAMBLE = """
# Add mypy-style generic params to a static type
# https://mypy.readthedocs.io/en/stable/generics.html#generic-type-aliases
TShape = typing.TypeVar('TShape')
TDtype = typing.TypeVar('TDtype')

"""
_STATIC_TYPES = ("TShape", "TDtype")


def generate_ndarray_stub() -> str:
    """
    Make a stub file based on the array interfaces that are available
    """

    import_strings = []
    type_names = [*_STATIC_TYPES]
    for arr in Interface.input_types():
        if arr.__module__ == "builtins":
            continue

        # Create import statements, saving aliased name of type if needed
        if arr.__module__.startswith("numpydantic") or arr.__module__ == "typing":
            type_name = str(arr) if arr.__module__ == "typing" else arr.__name__
            if arr.__module__ != "typing":
                import_strings.append(f"from {arr.__module__} import {type_name}")
        else:
            # since other packages could use the same name for an imported object
            # (eg dask and zarr both use an Array class)
            # we make an import alias from the module names to differentiate them
            # in the type annotation
            mod_name = "".join([a.capitalize() for a in arr.__module__.split(".")])
            type_name = mod_name + arr.__name__
            import_strings.append(
                f"from {arr.__module__} import {arr.__name__} " f"as {type_name}"
            )

        type_names.append(type_name)

    import_strings.extend(_BUILTIN_IMPORTS)
    import_strings = list(dict.fromkeys(import_strings))
    import_string = "\n".join(import_strings)

    class_union = " | ".join(type_names)
    ndarray_type = "NDArray = " + class_union

    stub_string = "\n".join([import_string, _PREAMBLE, ndarray_type])
    return stub_string


def update_ndarray_stub() -> None:
    """
    Update the ndarray.pyi string in the numpydantic file
    """
    from numpydantic import ndarray

    try:
        stub_string = generate_ndarray_stub()

        pyi_file = Path(ndarray.__file__).with_suffix(".pyi")
        with open(pyi_file, "w") as pyi:
            pyi.write(stub_string)
    except Exception as e:  # pragma: no cover
        warn(
            f"ndarray.pyi stub file could not be generated: {e}",
            category=ImportWarning,
            stacklevel=1,
        )



from datetime import datetime, timezone
from pathlib import Path
from typing import Optional, Tuple

import cv2
import dask.array as da
import h5py
import numpy as np
import zarr
from pydantic import BaseModel

from numpydantic.interface import (
    DaskInterface,
    H5ArrayPath,
    H5Interface,
    NumpyInterface,
    VideoInterface,
    ZarrArrayPath,
    ZarrInterface,
)
from numpydantic.testing.helpers import InterfaceCase
from numpydantic.types import DtypeType, NDArrayType


class NumpyCase(InterfaceCase):
    """In-memory numpy array"""

    interface = NumpyInterface

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> np.ndarray:
        if array is not None:
            return np.array(array, dtype=dtype)
        elif issubclass(dtype, BaseModel):
            return np.full(shape=shape, fill_value=dtype(x=1))
        elif len(shape) == 0:
            return dtype()
        else:
            return np.zeros(shape=shape, dtype=dtype)


class _HDF5MetaCase(InterfaceCase):
    """Base case for hdf5 cases"""

    interface = H5Interface

    @classmethod
    def skip(cls, shape: Tuple[int, ...], dtype: DtypeType) -> bool:
        return issubclass(dtype, BaseModel)


class HDF5Case(_HDF5MetaCase):
    """HDF5 Array"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[H5ArrayPath]:
        if cls.skip(shape, dtype):  # pragma: no cover
            return None

        hdf5_file = path / "h5f.h5"
        array_path = "/" + "_".join([str(s) for s in shape]) + "__" + dtype.__name__
        generator = np.random.default_rng()

        if array is not None:
            data = np.array(array, dtype=dtype)
        elif dtype is str:
            data = generator.random(shape).astype(bytes)
        elif dtype is np.str_:
            data = generator.random(shape).astype("S32")
        elif dtype is datetime:
            data = np.empty(shape, dtype="S32")
            data.fill(datetime.now(timezone.utc).isoformat().encode("utf-8"))
        else:
            data = generator.random(shape).astype(dtype)

        h5path = H5ArrayPath(hdf5_file, array_path)

        with h5py.File(hdf5_file, "w") as h5f:
            _ = h5f.create_dataset(array_path, data=data)
        return h5path


class HDF5CompoundCase(_HDF5MetaCase):
    """HDF5 Array with a fake compound dtype"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[H5ArrayPath]:
        if cls.skip(shape, dtype):  # pragma: no cover
            return None

        hdf5_file = path / "h5f.h5"
        array_path = "/" + "_".join([str(s) for s in shape]) + "__" + dtype.__name__
        if array is not None:
            data = np.array(array, dtype=dtype)
        elif dtype in (str, np.str_):
            dt = np.dtype([("data", np.dtype("S10")), ("extra", "i8")])
            data = np.array([("hey", 0)] * int(np.prod(shape)), dtype=dt).reshape(shape)
        elif dtype is datetime:
            dt = np.dtype([("data", np.dtype("S32")), ("extra", "i8")])
            data = np.array(
                [(datetime.now(timezone.utc).isoformat().encode("utf-8"), 0)]
                * int(np.prod(shape)),
                dtype=dt,
            ).reshape(shape)
        else:
            dt = np.dtype([("data", dtype), ("extra", "i8")])
            data = np.zeros(shape, dtype=dt)
        h5path = H5ArrayPath(hdf5_file, array_path, "data")

        with h5py.File(hdf5_file, "w") as h5f:
            _ = h5f.create_dataset(array_path, data=data)
        return h5path


class DaskCase(InterfaceCase):
    """In-memory dask array"""

    interface = DaskInterface

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> da.Array:
        if array is not None:
            return da.array(array, dtype=dtype)
        if issubclass(dtype, BaseModel):
            return da.full(shape=shape, fill_value=dtype(x=1), chunks=-1)
        else:
            return da.zeros(shape=shape, dtype=dtype, chunks=10)


class _ZarrMetaCase(InterfaceCase):
    """Shared classmethods for zarr cases"""

    interface = ZarrInterface

    @classmethod
    def skip(cls, shape: Tuple[int, ...], dtype: DtypeType) -> bool:
        return issubclass(dtype, BaseModel) or dtype is str


class ZarrCase(_ZarrMetaCase):
    """In-memory zarr array"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[zarr.Array]:
        if array is not None:
            return zarr.array(array, dtype=dtype, chunks=-1)
        else:
            return zarr.zeros(shape=shape, dtype=dtype)


class ZarrDirCase(_ZarrMetaCase):
    """On-disk zarr array"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[zarr.Array]:
        store = zarr.DirectoryStore(str(path / "array.zarr"))
        if array is not None:
            return zarr.array(array, dtype=dtype, store=store, chunks=-1)
        else:
            return zarr.zeros(shape=shape, dtype=dtype, store=store)


class ZarrZipCase(_ZarrMetaCase):
    """Zarr zip store"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[zarr.Array]:
        store = zarr.ZipStore(str(path / "array.zarr"), mode="w")
        if array is not None:
            return zarr.array(array, dtype=dtype, store=store, chunks=-1)
        else:
            return zarr.zeros(shape=shape, dtype=dtype, store=store)


class ZarrNestedCase(_ZarrMetaCase):
    """Nested zarr array"""

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> ZarrArrayPath:
        file = str(path / "nested.zarr")
        root = zarr.open(file, mode="w")
        subpath = "a/b/c"
        if array is not None:
            _ = root.array(subpath, array, dtype=dtype)
        else:
            _ = root.zeros(subpath, shape=shape, dtype=dtype)
        return ZarrArrayPath(file=file, path=subpath)


class VideoCase(InterfaceCase):
    """AVI video"""

    interface = VideoInterface

    @classmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10, 10, 3),
        dtype: DtypeType = np.uint8,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[Path]:
        if cls.skip(shape, dtype):  # pragma: no cover
            return None

        if array is not None:
            array = np.array(array, dtype=np.uint8)
            shape = array.shape

        is_color = len(shape) == 4
        frames = shape[0]
        frame_shape = shape[1:]

        video_path = path / "test.avi"
        writer = cv2.VideoWriter(
            str(video_path),
            cv2.VideoWriter_fourcc(*"RGBA"),  # raw video for testing purposes
            30,
            (frame_shape[1], frame_shape[0]),
            is_color,
        )
        for i in range(frames):
            if array is not None:
                frame = array[i]
            else:
                # make fresh array every time bc opencv eats them
                frame = np.full(frame_shape, fill_value=i, dtype=np.uint8)
            writer.write(frame)
        writer.release()
        return video_path

    @classmethod
    def skip(cls, shape: Tuple[int, ...], dtype: DtypeType) -> bool:
        """
        We really can only handle 4 dimensional cases in 8-bit rn lol

        .. todo::

            Fix shape/writing for grayscale videos

        """
        if len(shape) != 4:
            return True

        # if len(shape) < 3 or len(shape) > 4:
        #     return True
        if dtype not in (int, np.uint8):
            return True
        # if we have a color video (ie. shape == 4, needs to be RGB)
        if len(shape) == 4 and shape[3] != 3:
            return True



# ruff: noqa: F401
# ruff: noqa: I001
# ruff: noqa: D104

from numpydantic.ndarray import NDArray
from numpydantic.meta import update_ndarray_stub
from numpydantic.validation.shape import Shape

update_ndarray_stub()

__all__ = ["NDArray", "Shape"]



from pathlib import Path

import mypy.api
import pytest

from numpydantic import ndarray
from numpydantic.meta import update_ndarray_stub

MYPY_DIR = Path(__file__).parent / "data" / "mypy"


@pytest.fixture(scope="session", autouse=True)
def refresh_stubs():
    """Ensure no stale stubs"""
    # ensure all interfaces are imported
    from numpydantic import interface  # noqa: F401

    stub_file = Path(ndarray.__file__).with_suffix(".pyi")
    backup_file = stub_file.with_suffix(".pyi.bak")
    if stub_file.exists():
        stub_file.rename(backup_file)

    update_ndarray_stub()
    yield
    if backup_file.exists():
        stub_file.unlink(missing_ok=True)
        backup_file.rename(stub_file)


@pytest.mark.parametrize("test_file", MYPY_DIR.glob("*.py"))
def test_mypy(test_file: Path):
    """The mypy examples should pass static type checking"""
    res = mypy.api.run([str(test_file)])
    assert res == ("Success: no issues found in 1 source file\n", "", 0)



from dask.array.core import Array as DaskArrayCoreArray
from numpydantic.interface.hdf5 import H5ArrayPath
from numpydantic.interface.hdf5 import H5Proxy
from pathlib._local import Path as Pathlib_localPath
from cv2 import VideoCapture as Cv2VideoCapture
from numpydantic.interface.video import VideoProxy
from zarr.core import Array as ZarrCoreArray
from numpydantic.interface.zarr import ZarrArrayPath
from numpy import ndarray as Numpyndarray
import typing
import pathlib

# Add mypy-style generic params to a static type
# https://mypy.readthedocs.io/en/stable/generics.html#generic-type-aliases
TShape = typing.TypeVar('TShape')
TDtype = typing.TypeVar('TDtype')


NDArray = TShape | TDtype | DaskArrayCoreArray | H5ArrayPath | typing.Tuple[typing.Union[pathlib._local.Path, str], str] | H5Proxy | Pathlib_localPath | Cv2VideoCapture | VideoProxy | Pathlib_localPath | ZarrCoreArray | ZarrArrayPath | Numpyndarray


import json

import numpy as np
import pytest

from numpydantic.interface import ZarrInterface
from numpydantic.interface.zarr import ZarrArrayPath
from numpydantic.testing.cases import ZarrCase, ZarrDirCase, ZarrNestedCase, ZarrZipCase
from numpydantic.testing.helpers import InterfaceCase

pytestmark = pytest.mark.zarr


@pytest.fixture(
    params=[ZarrCase, ZarrZipCase, ZarrDirCase, ZarrNestedCase],
)
def zarr_case(request) -> InterfaceCase:
    return request.param


@pytest.fixture(
    params=[
        None,  # use the default store
        "dir_array",
        "zip_array",
        "nested_dir_array",
    ],
    ids=["MutableMapping", "DirectoryStore", "ZipStore", "NestedDirectoryStore"],
)
def store(request):
    if isinstance(request.param, str):
        return request.getfixturevalue(request.param)
    else:
        return request.param


def test_zarr_enabled():
    assert ZarrInterface.enabled()


def test_zarr_check(interface_cases, tmp_output_dir_func):
    """
    We should only use the zarr interface for zarr-like things
    """
    array = interface_cases.make_array(path=tmp_output_dir_func)
    if interface_cases.interface is ZarrInterface:
        assert ZarrInterface.check(array)
    else:
        assert not ZarrInterface.check(array)


@pytest.mark.shape
def test_zarr_shape(shape_cases, zarr_case):
    shape_cases.interface = zarr_case
    shape_cases.validate_case()


@pytest.mark.dtype
def test_zarr_dtype(dtype_cases, zarr_case):
    dtype_cases.interface = zarr_case
    if dtype_cases.skip():
        pytest.skip()
    dtype_cases.validate_case()


@pytest.mark.parametrize("array", ["zarr_nested_array", "zarr_array"])
def test_zarr_from_tuple(array, model_blank, request):
    """Should be able to do the same validation logic from tuples as an input"""
    array = request.getfixturevalue(array)
    if isinstance(array, ZarrArrayPath):
        _ = model_blank(array=(array.file, array.path))
    else:
        _ = model_blank(array=(array,))


def test_zarr_from_path(zarr_array, model_blank):
    """Should be able to just pass a path"""
    _ = model_blank(array=zarr_array)


def test_zarr_array_path_from_iterable(zarr_array):
    """Construct a zarr array path from some iterable!!!"""
    # from a single path
    apath = ZarrArrayPath.from_iterable((zarr_array,))
    assert apath.file == zarr_array
    assert apath.path is None

    inner_path = "/test/array"
    apath = ZarrArrayPath.from_iterable((zarr_array, inner_path))
    assert apath.file == zarr_array
    assert apath.path == inner_path


@pytest.mark.serialization
@pytest.mark.parametrize("dump_array", [True, False])
@pytest.mark.parametrize("roundtrip", [True, False])
def test_zarr_to_json(zarr_case, model_blank, roundtrip, dump_array, tmp_path):
    expected_fields = (
        "Type",
        "Data type",
        "Shape",
        "Chunk shape",
        "Compressor",
        "Store type",
        "hexdigest",
    )
    lol_array = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=int)

    array = zarr_case.make_array(array=lol_array, dtype=int, path=tmp_path)
    instance = model_blank(array=array)

    context = {"dump_array": dump_array}
    as_json = json.loads(
        instance.model_dump_json(round_trip=roundtrip, context=context)
    )["array"]

    if roundtrip:
        if dump_array:
            assert np.array_equal(as_json["value"], lol_array)
        else:
            if as_json.get("file", False):
                assert "array" not in as_json

        for field in expected_fields:
            assert field in as_json["info"]
        assert len(as_json["info"]["hexdigest"]) == 40

    else:
        assert np.array_equal(as_json, lol_array)



from typing import Any, Callable, Optional, Tuple, Type, Union

import numpy as np
import pytest
from pydantic import BaseModel, Field

from numpydantic import NDArray, Shape
from numpydantic.dtype import Number


@pytest.fixture(scope="function")
def array_model() -> (
    Callable[[Tuple[int, ...], Union[Type, np.dtype]], Type[BaseModel]]
):
    def _model(
        shape: Tuple[int, ...] = (10, 10), dtype: Union[Type, np.dtype] = float
    ) -> Type[BaseModel]:
        shape_str = ", ".join([str(s) for s in shape])

        class MyModel(BaseModel):
            array: NDArray[Shape[shape_str], dtype]

        return MyModel

    return _model


@pytest.fixture(scope="session")
def model_rgb() -> Type[BaseModel]:
    class RGB(BaseModel):
        array: Optional[
            Union[
                NDArray[Shape["* x, * y"], Number],
                NDArray[Shape["* x, * y, 3 r_g_b"], Number],
                NDArray[Shape["* x, * y, 3 r_g_b, 4 r_g_b_a"], Number],
            ]
        ] = Field(None)

    return RGB


@pytest.fixture(scope="session")
def model_blank() -> Type[BaseModel]:
    """A model with any shape and dtype"""

    class BlankModel(BaseModel):
        array: NDArray[Shape["*, ..."], Any]

    return BlankModel



from abc import ABC, abstractmethod
from collections.abc import Sequence
from functools import reduce
from itertools import product
from operator import ior
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
)

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, ValidationError, computed_field

from numpydantic import NDArray, Shape
from numpydantic.dtype import Float
from numpydantic.interface import Interface
from numpydantic.types import DtypeType, NDArrayType

if TYPE_CHECKING:
    from _pytest.mark.structures import MarkDecorator


class InterfaceCase(ABC):
    """
    An interface test helper that allows a given interface to generate and validate
    arrays in one of its formats.

    Each instance of "interface test case" should be considered one of the
    potentially multiple realizations of a given interface.
    If an interface has multiple formats (eg. zarr's different `store` s),
    then it should have several test helpers.
    """

    @property
    @abstractmethod
    def interface(self) -> Interface:
        """The interface that this helper is for"""

    @classmethod
    def array_from_case(
        cls, case: "ValidationCase", path: Optional[Path] = None
    ) -> Optional[NDArrayType]:
        """
        Generate an array from the given validation case.

        Returns ``None`` if an array can't be generated for a specific case.
        """
        return cls.make_array(shape=case.shape, dtype=case.dtype, path=path)

    @classmethod
    @abstractmethod
    def make_array(
        cls,
        shape: Tuple[int, ...] = (10, 10),
        dtype: DtypeType = float,
        path: Optional[Path] = None,
        array: Optional[NDArrayType] = None,
    ) -> Optional[NDArrayType]:
        """
        Make an array from a shape and dtype, and a path if needed

        Args:
            shape: shape of the array
            dtype: dtype of the array
            path: Path, if needed to generate on disk
            array: Rather than passing shape and dtype, pass a literal arraylike thing
        """

    @classmethod
    def validate_case(cls, case: "ValidationCase", path: Path) -> bool:
        """
        Validate a generated array against the annotation in the validation case.

        Kept in the InterfaceCase in case an interface has specific
        needs aside from just validating against a model, but typically left as is.

        If an array can't be generated for a given case, returns `None`
        so that the calling function can know to skip rather than fail the case.

        Raises exceptions if validation fails (or succeeds when it shouldn't)

        Args:
            case (ValidationCase): The validation case to validate.
            path (Path): Path to generate arrays into, if any.

        Returns:
            ``True`` if array is valid and was supposed to be,
                or invalid and wasn't supposed to be
        """
        import pytest

        array = cls.array_from_case(case, path)
        if array is None:
            pytest.skip()
        if case.passes:
            case.model(array=array)
            return True
        else:
            with pytest.raises(ValidationError):
                case.model(array=array)
            return True

    @classmethod
    def skip(cls, shape: Tuple[int, ...], dtype: DtypeType) -> bool:
        """
        Whether a given interface should be skipped for the case
        """
        # Assume an interface case is valid for all other cases
        return False


_a_shape_type = Tuple[Union[int, Literal["*"], Literal["..."]], ...]


class ValidationCase(BaseModel):
    """
    Test case for validating an array.

    Contains both the validating model and the parameterization for an array to
    test in a given interface
    """

    id: Optional[str] = None
    """
    String identifying the validation case
    """
    annotation_shape: Union[
        Tuple[Union[int, str], ...],
        Tuple[Tuple[Union[int, str], ...], ...],
    ] = (10, 10, "*", "*")
    """
    Shape to use in computed annotation used to validate against
    """
    annotation_dtype: Union[DtypeType, Sequence[DtypeType]] = Float
    """
    Dtype to use in computed annotation used to validate against
    """
    shape: Tuple[int, ...] = (10, 10, 2, 2)
    """Shape of the array to validate"""
    dtype: Union[Type, np.dtype] = float
    """Dtype of the array to validate"""
    passes: bool = False
    """Whether the validation should pass or not"""
    interface: Optional[Type[InterfaceCase]] = None
    """The interface test case to generate and validate the array with"""
    path: Optional[Path] = None
    """The path to generate arrays into, if any."""
    marks: set[str] = Field(default_factory=set)
    """pytest marks to set for this test case"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @computed_field()
    def annotation(self) -> NDArray:
        """
        Annotation used in the model we validate against
        """
        # make a union type if we need to
        shape_union = all(
            isinstance(s, Sequence) and not isinstance(s, str)
            for s in self.annotation_shape
        )
        dtype_union = isinstance(self.annotation_dtype, Sequence) and all(
            isinstance(s, Sequence) for s in self.annotation_dtype
        )
        if shape_union or dtype_union:
            shape_iter = (
                self.annotation_shape if shape_union else [self.annotation_shape]
            )
            dtype_iter = (
                self.annotation_dtype if dtype_union else [self.annotation_dtype]
            )
            annotations: List[type] = []
            for shape, dtype in product(shape_iter, dtype_iter):
                shape_str = ", ".join([str(i) for i in shape])
                annotations.append(NDArray[Shape[shape_str], dtype])
            return Union[tuple(annotations)]

        else:
            shape_str = ", ".join([str(i) for i in self.annotation_shape])
            return NDArray[Shape[shape_str], self.annotation_dtype]

    @computed_field()
    def model(self) -> Type[BaseModel]:
        """A model with a field ``array`` with the given annotation"""
        annotation = self.annotation

        class Model(BaseModel):
            array: annotation

        return Model

    @property
    def pytest_marks(self) -> list["MarkDecorator"]:
        """
        Instantiated pytest marks from :attr:`.ValidationCase.marks`
        plus the interface name.
        """
        import pytest

        marks = self.marks.copy()
        if self.interface is not None:
            marks.add(self.interface.interface.name)
        return [getattr(pytest.mark, m) for m in marks]

    def validate_case(self, path: Optional[Path] = None) -> bool:
        """
        Whether the generated array correctly validated against the annotation,
        given the interface

        Args:
            path (:class:`pathlib.Path`): Directory to generate array into, if on disk.

        Raises:
            ValueError: if an ``interface`` is missing
        """
        if self.interface is None:  # pragma: no cover
            raise ValueError("Missing an interface")
        if path is None:
            if self.path:
                path = self.path
            else:  # pragma: no cover
                raise ValueError("Missing a path to generate arrays into")

        return self.interface.validate_case(self, path)

    def array(self, path: Path) -> NDArrayType:
        """Generate an array for the validation case if we have an interface to do so"""
        if self.interface is None:  # pragma: no cover
            raise ValueError("Missing an interface")
        if path is None:  # pragma: no cover
            if self.path:
                path = self.path
            else:
                raise ValueError("Missing a path to generate arrays into")

        return self.interface.array_from_case(self, path)

    def merge(
        self, other: Union["ValidationCase", Sequence["ValidationCase"]]
    ) -> "ValidationCase":
        """
        Merge two validation cases

        Dump both, excluding any unset fields, and merge, preferring `other`.

        ``valid`` is ``True`` if and only if it is ``True`` in both.
        """
        if isinstance(other, Sequence):
            return merge_cases(self, *other)
        else:
            return merge_cases(self, other)

    def skip(self) -> bool:
        """
        Whether this case should be skipped
        (eg. due to the interface case being incompatible
        with the requested dtype or shape)
        """
        return bool(
            self.interface is not None and self.interface.skip(self.shape, self.dtype)
        )


def merge_cases(*args: ValidationCase) -> ValidationCase:
    """
    Merge multiple validation cases
    """
    if len(args) == 1:  # pragma: no cover
        return args[0]

    dumped = [
        m.model_dump(
            exclude_unset=True, exclude={"model", "annotation", "pytest_marks"}
        )
        for m in args
    ]

    # self_dump = self.model_dump(exclude_unset=True)
    # other_dump = other.model_dump(exclude_unset=True)

    # dumps might not have set `passes`, use only the ones that have
    passes = [v.get("passes") for v in dumped if "passes" in v]
    passes = all(passes)

    # combine ids if present
    ids = "-".join([str(v.get("id")) for v in dumped if "id" in v])

    # merge dicts
    merged = reduce(ior, dumped, {})
    merged["passes"] = passes
    merged["id"] = ids
    merged["marks"] = set().union(*[v.get("marks", set()) for v in dumped])
    return ValidationCase.model_construct(**merged)


def merged_product(
    *args: Sequence[ValidationCase], conditions: dict = None
) -> list[ValidationCase]:
    """
    Generator for the product of the iterators of validation cases,
    merging each tuple, and respecting if they should be :meth:`.ValidationCase.skip`
    or not.

    Examples:

        .. code-block:: python

            shape_cases = [
                ValidationCase(shape=(10, 10, 10), passes=True, id="valid shape"),
                ValidationCase(shape=(10, 10), passes=False, id="missing dimension"),
            ]
            dtype_cases = [
                ValidationCase(dtype=float, passes=True, id="float"),
                ValidationCase(dtype=int, passes=False, id="int"),
            ]

            iterator = merged_product(shape_cases, dtype_cases))
            next(iterator)
            # ValidationCase(
            #     shape=(10, 10, 10),
            #     dtype=float,
            #     passes=True,
            #     id="valid shape-float"
            # )
            next(iterator)
            # ValidationCase(
            #     shape=(10, 10, 10),
            #     dtype=int,
            #     passes=False,
            #     id="valid shape-int"
            # )


    """
    iterator = product(*args)
    cases = []
    for case_tuple in iterator:
        case = merge_cases(*case_tuple)
        if case.skip():
            continue
        if conditions:
            matching = all([getattr(case, k, None) == v for k, v in conditions.items()])
            if not matching:
                continue
        cases.append(case)
    return cases



"""
Extension of nptyping NDArray for pydantic that allows for JSON-Schema serialization

.. note::

    This module should *only* have the :class:`.NDArray` class in it, because the
    type stub ``ndarray.pyi`` is only created for :class:`.NDArray` . Otherwise,
    type checkers will complain about using any helper functions elsewhere -
    those all belong in :mod:`numpydantic.schema` .

    Keeping with nptyping's style, NDArrayMeta is in this module even if it's
    excluded from the type stub.

"""

from typing import TYPE_CHECKING, Any, Literal, Tuple, get_origin

import numpy as np
from pydantic import GetJsonSchemaHandler
from pydantic_core import core_schema

from numpydantic.dtype import DType
from numpydantic.exceptions import InterfaceError
from numpydantic.interface import Interface
from numpydantic.maps import python_to_nptyping
from numpydantic.schema import (
    get_validate_interface,
    make_json_schema,
)
from numpydantic.serialization import jsonize_array
from numpydantic.types import DtypeType, NDArrayType, ShapeType
from numpydantic.validation.dtype import is_union
from numpydantic.vendor.nptyping.error import InvalidArgumentsError
from numpydantic.vendor.nptyping.ndarray import NDArrayMeta as _NDArrayMeta
from numpydantic.vendor.nptyping.nptyping_type import NPTypingType
from numpydantic.vendor.nptyping.structure import Structure
from numpydantic.vendor.nptyping.structure_expression import check_type_names
from numpydantic.vendor.nptyping.typing_ import (
    dtype_per_name,
)

if TYPE_CHECKING:  # pragma: no cover
    from nptyping.base_meta_classes import SubscriptableMeta
    from pydantic._internal._schema_generation_shared import (
        CallbackGetCoreSchemaHandler,
    )

    from numpydantic import Shape


class NDArrayMeta(_NDArrayMeta, implementation="NDArray"):
    """
    Hooking into nptyping's array metaclass to override methods pending
    completion of the transition away from nptyping
    """

    if TYPE_CHECKING:  # pragma: no cover
        __getitem__ = SubscriptableMeta.__getitem__

    def __call__(cls, val: NDArrayType) -> NDArrayType:
        """Call ndarray as a validator function"""
        return get_validate_interface(cls.__args__[0], cls.__args__[1])(val)

    def __instancecheck__(self, instance: Any):
        """
        Extended type checking that determines whether

        1) the ``type`` of the given instance is one of those in
            :meth:`.Interface.input_types`

        but also

        2) it satisfies the constraints set on the :class:`.NDArray` annotation

        Args:
            instance (:class:`typing.Any`): Thing to check!

        Returns:
            bool: ``True`` if matches constraints, ``False`` otherwise.
        """
        shape, dtype = self.__args__
        try:
            interface_cls = Interface.match(instance, fast=True)
            interface = interface_cls(shape, dtype)
            _ = interface.validate(instance)
            return True
        except InterfaceError:
            return False

    def _is_literal_like(cls, item: Any) -> bool:
        """
        Changes from nptyping:
        - doesn't just ducktype for literal but actually, yno, checks for being literal
        """
        return get_origin(item) is Literal

    def _get_shape(cls, dtype_candidate: Any) -> "Shape":
        """
        Override of base method to use our local definition of shape
        """
        from numpydantic.validation.shape import Shape

        if dtype_candidate is Any or dtype_candidate is Shape:
            shape = Any
        elif issubclass(dtype_candidate, Shape):
            shape = dtype_candidate
        elif cls._is_literal_like(dtype_candidate):
            shape_expression = dtype_candidate.__args__[0]
            shape = Shape[shape_expression]
        else:
            raise InvalidArgumentsError(
                f"Unexpected argument '{dtype_candidate}', expecting"
                " Shape[<ShapeExpression>]"
                " or Literal[<ShapeExpression>]"
                " or typing.Any."
            )
        return shape

    def _get_dtype(cls, dtype_candidate: Any) -> DType:
        """
        Override of base _get_dtype method to allow for compound tuple types
        """
        if dtype_candidate in python_to_nptyping:
            dtype_candidate = python_to_nptyping[dtype_candidate]
        is_dtype = isinstance(dtype_candidate, type) and issubclass(
            dtype_candidate, np.generic
        )

        if dtype_candidate is Any:
            dtype = Any
        elif is_dtype or is_union(dtype_candidate):
            dtype = dtype_candidate
        elif issubclass(dtype_candidate, Structure):  # pragma: no cover
            dtype = dtype_candidate
            check_type_names(dtype, dtype_per_name)
        elif cls._is_literal_like(dtype_candidate):  # pragma: no cover
            structure_expression = dtype_candidate.__args__[0]
            dtype = Structure[structure_expression]
            check_type_names(dtype, dtype_per_name)
        elif isinstance(dtype_candidate, tuple):  # pragma: no cover
            dtype = tuple([cls._get_dtype(dt) for dt in dtype_candidate])
        else:
            # arbitrary dtype - allow failure elsewhere :)
            dtype = dtype_candidate

        return dtype

    def _dtype_to_str(cls, dtype: Any) -> str:
        if dtype is Any:
            result = "Any"
        elif issubclass(dtype, Structure):
            result = str(dtype)
        elif isinstance(dtype, tuple):
            result = ", ".join([str(dt) for dt in dtype])
        else:
            result = str(dtype)
        return result


class NDArray(NPTypingType, metaclass=NDArrayMeta):
    """
    Constrained array type allowing npytyping syntax for dtype and shape validation
    and serialization.

    This class is not intended to be instantiable, and support for static type
    checking is limited,
    it implements the ``__get_pydantic_core_schema__`` method to invoke
    the relevant :ref:`interface <Interfaces>` for validation and serialization.

    It is callable, however, which validates and attempts to coerce input to a
    supported array type.
    There is no such thing as an "NDArray instance," but one can think of it
    as a validating passthrough callable.

    References:
        - https://docs.pydantic.dev/latest/usage/types/custom/#handling-third-party-types
    """

    __args__: Tuple[ShapeType, DtypeType] = (Any, Any)

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: "NDArray",
        _handler: "CallbackGetCoreSchemaHandler",
    ) -> core_schema.CoreSchema:
        shape, dtype = _source_type.__args__
        shape: ShapeType
        dtype: DtypeType

        # make core schema for json schema, store it and any model definitions
        # so that we can use them when rendering json schema
        json_schema = make_json_schema(shape, dtype, _handler)

        return core_schema.with_info_plain_validator_function(
            get_validate_interface(shape, dtype),
            serialization=core_schema.plain_serializer_function_ser_schema(
                jsonize_array, when_used="json", info_arg=True
            ),
            metadata=json_schema,
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> core_schema.JsonSchema:
        shape, dtype = cls.__args__
        json_schema = handler(schema["metadata"])
        json_schema = handler.resolve_ref_schema(json_schema)

        if (
            not isinstance(dtype, tuple)
            and dtype.__module__
            not in (
                "builtins",
                "typing",
                "types",
            )
            and hasattr(dtype, "__name__")
        ):
            json_schema["dtype"] = ".".join([dtype.__module__, dtype.__name__])

        return json_schema



"""
MIT License

Copyright (c) 2023 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import string
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Union,
)

from numpydantic.vendor.nptyping.error import InvalidShapeError
from numpydantic.vendor.nptyping.typing_ import ShapeExpression, ShapeTuple

if TYPE_CHECKING:
    from nptyping.shape import Shape  # pragma: no cover


@lru_cache
def check_shape(shape: ShapeTuple, target: "Shape") -> bool:
    """
    Check whether the given shape corresponds to the given shape_expression.
    :param shape: the shape in question.
    :param target: the shape expression to which shape is tested.
    :return: True if the given shape corresponds to shape_expression.
    """
    target_shape = _handle_ellipsis(shape, target.prepared_args)
    return _check_dimensions_against_shape(shape, target_shape)


def validate_shape_expression(shape_expression: Union[ShapeExpression, Any]) -> None:
    """
    Validate shape_expression and raise an InvalidShapeError if it is not
    considered valid.
    :param shape_expression: the shape expression to validate.
    :return: None.
    """
    shape_expression_no_quotes = shape_expression.replace("'", "").replace('"', "")
    if shape_expression is not Any and not re.match(
        _REGEX_SHAPE_EXPRESSION, shape_expression_no_quotes
    ):
        raise InvalidShapeError(
            f"'{shape_expression}' is not a valid shape expression."
        )


def normalize_shape_expression(shape_expression: ShapeExpression) -> ShapeExpression:
    """
    Normalize the given shape expression, e.g. by removing whitespaces, making
    similar expressions look the same.
    :param shape_expression: the shape expression that is to be normalized.
    :return: a normalized shape expression.
    """
    shape_expression = shape_expression.replace("'", "").replace('"', "")
    # Replace whitespaces right before labels with $.
    shape_expression = re.sub(rf"\s*{_REGEX_LABEL}", r"$\1", shape_expression)
    # Let all commas be followed by a $.
    shape_expression = shape_expression.replace(",", ",$")
    # Remove all whitespaces left.
    shape_expression = re.sub(r"\s*", "", shape_expression)
    # Remove $ right after a bracket.
    shape_expression = re.sub(r"\[\$+", "[", shape_expression)
    # Replace $ with a single space.
    shape_expression = re.sub(r"\$+", " ", shape_expression)
    return shape_expression


def get_dimensions(shape_expression: str) -> List[str]:
    """
    Find all "break downs" (the parts between brackets) in a shape expressions
    and replace them with mere dimension sizes.

    :param shape_expression: the shape expression that gets the break downs replaced.
    :return: a list of dimensions without break downs.
    """
    shape_expression_without_breakdowns = shape_expression
    for dim_breakdown in re.findall(
        r"(\[[^\]]+\])", shape_expression_without_breakdowns
    ):
        dim_size = len(dim_breakdown.split(","))
        shape_expression_without_breakdowns = (
            shape_expression_without_breakdowns.replace(dim_breakdown, str(dim_size))
        )
    return shape_expression_without_breakdowns.split(",")


def remove_labels(dimensions: List[str]) -> List[str]:
    """
    Remove all labels (words that start with a lowercase).

    :param dimensions: a list of dimensions.
    :return: a copy of the given list without labels.
    """
    return [re.sub(r"\b[a-z]\w*", "", dim).strip() for dim in dimensions]


def _check_dimensions_against_shape(shape: ShapeTuple, target: List[str]) -> bool:
    # Walk through the shape and test them against the given target,
    # taking into consideration variables, wildcards, etc.

    if len(shape) != len(target):
        return False
    shape_as_strings = (str(dim) for dim in shape)
    variables: Dict[str, str] = {}
    for dim, target_dim in zip(shape_as_strings, target):
        if _is_wildcard(target_dim) or _is_assignable_var(dim, target_dim, variables):
            continue
        if dim != target_dim:
            return False
    return True


def _handle_ellipsis(shape: ShapeTuple, target: List[str]) -> List[str]:
    # Let the ellipsis allows for any number of dimensions by replacing the
    # ellipsis with the dimension size repeated the number of times that
    # corresponds to the shape of the instance.
    if target[-1] == "...":
        dim_to_repeat = target[-2]
        target = target[0:-1]
        if len(shape) > len(target):
            difference = len(shape) - len(target)
            target += difference * [dim_to_repeat]
    return target


def _is_assignable_var(dim: str, target_dim: str, variables: Dict[str, str]) -> bool:
    # Return whether target_dim is a variable and can be assigned with dim.
    return _is_variable(target_dim) and _can_assign_variable(dim, target_dim, variables)


def _is_variable(dim: str) -> bool:
    # Return whether dim is a variable.
    return dim[0] in string.ascii_uppercase


def _can_assign_variable(dim: str, target_dim: str, variables: Dict[str, str]) -> bool:
    # Check and assign a variable.
    assignable = variables.get(target_dim) in (None, dim)
    variables[target_dim] = dim
    return assignable


def _is_wildcard(dim: str) -> bool:
    # Return whether dim is a wildcard (i.e. the character that takes any
    # dimension size).
    return dim == "*"


_REGEX_SEPARATOR = r"(\s*,\s*)"
_REGEX_DIMENSION_SIZE = r"(\s*[0-9]+\s*)"
_REGEX_VARIABLE = r"(\s*\b[A-Z]\w*\s*)"
_REGEX_LABEL = r"(\s*\b[a-z]\w*\s*)"
_REGEX_LABELS = rf"({_REGEX_LABEL}({_REGEX_SEPARATOR}{_REGEX_LABEL})*)"
_REGEX_WILDCARD = r"(\s*\*\s*)"
_REGEX_DIMENSION_BREAKDOWN = rf"(\s*\[{_REGEX_LABELS}\]\s*)"
_REGEX_DIMENSION = (
    rf"({_REGEX_DIMENSION_SIZE}"
    rf"|{_REGEX_VARIABLE}"
    rf"|{_REGEX_WILDCARD}"
    rf"|{_REGEX_DIMENSION_BREAKDOWN})"
)
_REGEX_DIMENSION_WITH_LABEL = rf"({_REGEX_DIMENSION}(\s+{_REGEX_LABEL})*)"
_REGEX_DIMENSIONS = (
    rf"{_REGEX_DIMENSION_WITH_LABEL}({_REGEX_SEPARATOR}{_REGEX_DIMENSION_WITH_LABEL})*"
)
_REGEX_DIMENSIONS_ELLIPSIS = rf"({_REGEX_DIMENSIONS}{_REGEX_SEPARATOR}\.\.\.\s*)"
_REGEX_SHAPE_EXPRESSION = rf"^({_REGEX_DIMENSIONS}|{_REGEX_DIMENSIONS_ELLIPSIS})$"



"""
Helper functions for validation of dtype.

For literal dtypes intended for use by end-users, see :mod:`numpydantic.dtype`
"""

import sys
from typing import Any, Union, get_args, get_origin

import numpy as np

from numpydantic.types import DtypeType

if sys.version_info >= (3, 10):
    from types import UnionType
else:
    UnionType = None


def validate_dtype(dtype: Any, target: DtypeType) -> bool:
    """
    Validate a dtype against the target dtype.

    If `dtype` or `target` are `Any`, validation passes trivially.
    The `dtype` may be `Any` when the dtype can't be determined,
    but failure to determine dtype shouldn't be fatal
    (e.g. BaseModel dtypes for empty arrays).

    Args:
        dtype: The dtype to validate
        target (:class:`.DtypeType`): The target dtype

    Returns:
        bool: ``True`` if valid, ``False`` otherwise
    """
    if target is Any or dtype is Any:
        return True

    if isinstance(target, tuple):
        valid = any(validate_dtype(dtype, target_dt) for target_dt in target)
    elif is_union(target):
        valid = any(
            [validate_dtype(dtype, target_dt) for target_dt in get_args(target)]
        )
    elif target is np.str_:
        valid = getattr(dtype, "type", None) in (np.str_, str) or dtype in (
            np.str_,
            str,
        )
    else:
        # try to match as any subclass, if target is a class
        try:
            valid = issubclass(dtype, target)
        except TypeError:
            # error expected if dtype or target is not a class
            # main type check - directly test dtype identity
            valid = dtype == target or getattr(dtype, "type", None) == target

    return valid


def is_union(dtype: DtypeType) -> bool:
    """
    Check if a dtype is a union
    """
    if UnionType is None:
        return get_origin(dtype) is Union
    else:
        return get_origin(dtype) in (Union, UnionType)



"""
MIT License

Copyright (c) 2023 Ramon Hagenaars

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import inspect
from abc import ABC
from typing import Any, Tuple

import numpy as np

from numpydantic.vendor.nptyping import InvalidArgumentsError
from numpydantic.vendor.nptyping.base_meta_classes import (
    FinalMeta,
    InconstructableMeta,
    MaybeCheckableMeta,
    PrintableMeta,
    SubscriptableMeta,
)
from numpydantic.vendor.nptyping.error import DependencyError
from numpydantic.vendor.nptyping.nptyping_type import NPTypingType
from numpydantic.vendor.nptyping.pandas_.typing_ import dtype_per_name
from numpydantic.vendor.nptyping.structure import Structure
from numpydantic.vendor.nptyping.structure_expression import check_structure

try:
    import pandas as pd
except ImportError:  # pragma: no cover
    pd = None  # type: ignore[misc, assignment]


class DataFrameMeta(
    SubscriptableMeta,
    InconstructableMeta,
    FinalMeta,
    MaybeCheckableMeta,
    PrintableMeta,
    implementation="DataFrame",
):
    """
    Metaclass that is coupled to nptyping.DataFrame. It contains all actual logic
    such as instance checking.
    """

    __args__: Tuple[Structure]
    _parameterized: bool

    def __instancecheck__(  # pylint: disable=bad-mcs-method-argument
        self, instance: Any
    ) -> bool:
        structure = self.__args__[0]

        if pd is None:
            raise DependencyError(  # pragma: no cover
                "Pandas needs to be installed for instance checking. Use `pip "
                "install nptyping[pandas]` or `pip install nptyping[complete]`"
            )

        if not isinstance(instance, pd.DataFrame):
            return False

        if structure is Any:
            return True

        structured_dtype = np.dtype(
            [(column, dtype.str) for column, dtype in instance.dtypes.items()]
        )
        return check_structure(structured_dtype, structure, dtype_per_name)

    def _get_item(cls, item: Any) -> Tuple[Structure]:
        if item is Any:
            return (Any,)
        cls._check_item(item)
        return (Structure[item.__args__[0]],)

    def __str__(cls) -> str:
        structure = cls.__args__[0]
        structure_str = "Any" if structure is Any else structure.__args__[0]
        return f"{cls.__name__}[{structure_str}]"

    def __repr__(cls) -> str:
        structure = cls.__args__[0]
        structure_str = "Any" if structure is Any else structure
        return f"{cls.__name__}[{structure_str}]"

    @property
    def __module__(cls) -> str:
        return cls._get_module(inspect.currentframe(), "nptyping.ndarray")

    def _check_item(cls, item: Any) -> None:
        # Check if the item is what we expect and raise if it is not.
        if not hasattr(item, "__args__"):
            raise InvalidArgumentsError(f"Unexpected argument of type {type(item)}.")


class DataFrame(NPTypingType, ABC, metaclass=DataFrameMeta):
    """
    An nptyping equivalent of pandas DataFrame.

    ## No arguments means a DataFrame of any structure.
    >>> DataFrame
    DataFrame[Any]

    ## You can use Structure Expression.
    >>> from nptyping import DataFrame, Structure
    >>> DataFrame[Structure["x: Int, y: Int"]]
    DataFrame[Structure['[x, y]: Int']]

    ## Instance checking can be done and the structure is also checked.
    >>> import pandas as pd
    >>> df = pd.DataFrame({'x': [1, 2, 3], 'y': [4., 5., 6.]})
    >>> isinstance(df, DataFrame[Structure['x: Int, y: Float']])
    True
    >>> isinstance(df, DataFrame[Structure['x: Float, y: Int']])
    False

    """

    __args__ = (Any,)



import shutil
from _warnings import warn
from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def tmp_output_dir(request: pytest.FixtureRequest) -> Path:
    path = Path(__file__).parents[1].resolve() / "__tmp__"
    if path.exists():
        shutil.rmtree(str(path))
    path.mkdir()

    yield path

    if not request.config.getvalue("--with-output"):
        try:
            shutil.rmtree(str(path))
        except PermissionError as e:
            # sporadic error on windows machines...
            warn(
                "Temporary directory could not be removed due to a permissions error: "
                f"\n{str(e)}"
            )


@pytest.fixture(scope="function")
def tmp_output_dir_func(tmp_output_dir, request: pytest.FixtureRequest) -> Path:
    """
    tmp output dir that gets cleared between every function
    cleans at the start rather than at cleanup in case the output is to be inspected
    """
    subpath = tmp_output_dir / f"__tmpfunc_{request.node.name}__"
    if subpath.exists():
        shutil.rmtree(str(subpath))
    subpath.mkdir()
    return subpath


@pytest.fixture(scope="module")
def tmp_output_dir_mod(tmp_output_dir, request: pytest.FixtureRequest) -> Path:
    """
    tmp output dir that gets cleared between every function
    cleans at the start rather than at cleanup in case the output is to be inspected
    """
    subpath = tmp_output_dir / f"__tmpmod_{request.module}__"
    if subpath.exists():
        shutil.rmtree(str(subpath))
    subpath.mkdir()
    return subpath
